"""Abstract backend."""

import logging

from abc import ABCMeta, abstractmethod

import pyparsing

from cumin import CuminError, nodeset


class InvalidQueryError(CuminError):
    """Custom exception class for invalid queries."""


class BaseQuery(metaclass=ABCMeta):
    """Query abstract class.

    All backends query classes must inherit, directly or indirectly, from this one.
    """

    grammar = pyparsing.NoMatch()  # This grammar will never match.
    """:py:class:`pyparsing.ParserElement`: derived classes must define their own pyparsing grammar and set this class
    attribute accordingly."""

    def __init__(self, config):
        """Query constructor.

        Arguments:
            config (dict): a dictionary with the parsed configuration file.

        """
        self.config = config
        self.logger = logging.getLogger('.'.join((self.__module__, self.__class__.__name__)))
        self.logger.trace('Backend %s created with config: %s', type(self).__name__, config)

    def execute(self, query_string):
        """Build and execute the query, return the NodeSet of FQDN hostnames that matches.

        Arguments:
            query_string (str): the query string to be parsed and executed.

        Returns:
            ClusterShell.NodeSet.NodeSet: with the FQDNs of the matching hosts.

        """
        self._build(query_string)
        return self._execute()

    @abstractmethod
    def _execute(self):
        """Execute the already parsed query and return the NodeSet of FQDN hostnames that matches.

        Returns:
            ClusterShell.NodeSet.NodeSet: with the FQDNs of the matching hosts.

        """

    @abstractmethod
    def _parse_token(self, token):
        """Recursively interpret the tokens returned by the grammar parsing.

        Arguments:
            token (pyparsing.ParseResults): a single token returned by the grammar parsing.

        """

    def _build(self, query_string):
        """Parse the query string according to the grammar and build the query for later execution.

        Arguments:
            query_string (str): the query string to be parsed.

        """
        self.logger.trace('Parsing query: %s', query_string)
        parsed = self.grammar.parseString(query_string.strip(), parseAll=True)
        self.logger.trace('Parsed query: %s', parsed)
        for token in parsed:
            self._parse_token(token)


class BaseQueryAggregator(BaseQuery):
    """Query aggregator abstract class.

    Add to :py:class:`cumin.backends.BaseQuery` the capability of aggregating query subgroups and sub tokens into a
    unified result using common boolean operators for sets: ``and``, ``or``, ``and not`` and ``xor``.
    The class has a stack-like structure that must be populated by the derived classes while building the query.
    On execution the stack is traversed and the results are aggreagated together based on subgroups and boolean
    operators.
    """

    def __init__(self, config):
        """Query aggregator constructor, initialize the stack.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery.__init__`.
        """
        super().__init__(config)

        self.stack = None
        self.stack_pointer = None

    def _build(self, query_string):
        """Override parent method to reset the stack and log it.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._build`.
        """
        self.stack = self._get_stack_element()
        self.stack_pointer = self.stack
        super()._build(query_string)
        self.logger.trace('Query stack: %s', self.stack)

    def _execute(self):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._execute`.
        """
        hosts = nodeset()
        self._loop_stack(hosts, self.stack)  # The hosts NodeSet is updated in place while looping the stack
        self.logger.debug('Found %d hosts', len(hosts))

        return hosts

    def _open_subgroup(self):
        """Handle subgroup opening."""
        element = self._get_stack_element()
        element['parent'] = self.stack_pointer
        self.stack_pointer['children'].append(element)
        self.stack_pointer = element

    def _close_subgroup(self):
        """Handle subgroup closing."""
        self.stack_pointer = self.stack_pointer['parent']

    @abstractmethod
    def _parse_token(self, token):
        """Re-define abstract method from parent abstract class.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._parse_token`.
        """

    @staticmethod
    def _get_stack_element():
        """Return an empty stack element.

        Returns:
            dict: the dictionary with an empty stack element.

        """
        return {'hosts': None, 'children': [], 'parent': None, 'bool': None}

    def _loop_stack(self, hosts, stack_element):
        """Loop the stack generated while parsing the query and aggregate the results.

        Arguments:
            hosts (ClusterShell.NodeSet.NodeSet): the hosts to be updated with the current stack element results. This
                object is updated in place by reference.
            stack_element (dict): the stack element to iterate.

        """
        if stack_element['hosts'] is None:
            element_hosts = nodeset()
            for child in stack_element['children']:
                self._loop_stack(element_hosts, child)
        else:
            element_hosts = stack_element['hosts']

        self._aggregate_hosts(hosts, element_hosts, stack_element['bool'])

    def _aggregate_hosts(self, hosts, element_hosts, bool_operator):
        """Aggregate hosts according to their boolean operator.

        Arguments:
            hosts (ClusterShell.NodeSet.NodeSet): the hosts to update with the results in ``element_hosts`` according
                to the ``bool_operator``. This object is updated in place by reference.
            element_hosts (ClusterShell.NodeSet.NodeSet): the additional hosts to aggregate to the results based on the
                ``bool_operator``.
            bool_operator (str, None): the boolean operator to apply while aggregating the two NodeSet. It must be
                :py:data:`None` when adding the first hosts.

        """
        self.logger.trace("Aggregating: %s | %s | %s", hosts, bool_operator, element_hosts)

        # This should never happen
        if (bool_operator is None and hosts) or (bool_operator is not None and not hosts):  # pragma: no cover
            raise InvalidQueryError("Unexpected boolean operator '{boolean}' with hosts '{hosts}'".format(
                boolean=bool_operator, hosts=hosts))

        if bool_operator is None or bool_operator == 'or':
            hosts |= element_hosts
        elif bool_operator == 'and':
            hosts &= element_hosts
        elif bool_operator == 'and not':
            hosts -= element_hosts
        elif bool_operator == 'xor':
            hosts ^= element_hosts
        else:  # pragma: no cover - this should never happen
            raise InvalidQueryError("Invalid bool operator '{boolean}' found, one of and|and not|or expected".format(
                boolean=bool_operator))



"""Transport factory."""

import importlib

from cumin import CuminError


class Transport:
    """Transport factory class.

    The transport layer is the one used to convey the commands to be executed into the selected hosts. The transport
    abstraction allow to specify a mode to choose the execution plan, an event handler class and a success threshold.
    Those can be used by the chosen transport to customize the behavior of the execution plan.

    All the transports share a common interface that is defined in the :py:class:`cumin.transports.BaseWorker` class
    and they are instantiated through the :py:class:`cumin.transport.Transport` factory class. Each transport module
    need to define a ``worker_class`` module variable that is a pointer to the transport class for dynamic
    instantiation.
    """

    @staticmethod
    def new(config, target):
        """Create a transport worker class based on the configuration (`factory`).

        Arguments:
            config (dict): the configuration dictionary.
            target (cumin.transports.Target): a Target instance.

        Returns:
            BaseWorker: the created worker instance for the configured transport.

        Raises:
            cumin.CuminError: if the configuration is missing the required ``transport`` key.
            exceptions.ImportError: if unable to import the transport module.
            exceptions.AttributeError: if the transport module is missing the required ``worker_class`` attribute.

        """
        if 'transport' not in config:
            raise CuminError("Missing required parameter 'transport' in the configuration dictionary")

        try:
            module = importlib.import_module('cumin.transports.{transport}'.format(transport=config['transport']))
            return module.worker_class(config, target)
        except (AttributeError, ImportError) as e:
            raise CuminError("Unable to load worker class for transport '{transport}': {msg}".format(
                transport=config['transport'], msg=repr(e))) from e



"""Backends basic grammar tests."""
import importlib
import os
import pkgutil

import pytest

from cumin import backends
from cumin.tests import get_fixture


BACKENDS = [name for _, name, ispkg in pkgutil.iter_modules(backends.__path__) if not ispkg]
BASE_PATH = os.path.join('backends', 'grammars')


@pytest.mark.parametrize('backend_name', BACKENDS)
def test_valid_grammars(backend_name):
    """Run quick pyparsing test over valid grammar strings for each backend that has the appropriate fixture."""
    try:
        backend = importlib.import_module('cumin.backends.{backend}'.format(backend=backend_name))
    except ImportError:
        return  # Backend not available

    results = backend.grammar().runTests(
        get_fixture(os.path.join(BASE_PATH, '{backend}_valid.txt'.format(backend=backend_name)), as_string=True))
    assert results[0]


@pytest.mark.parametrize('backend_name', BACKENDS)
def test_invalid_grammars(backend_name):
    """Run quick pyparsing test over invalid grammar strings for each backend that has the appropriate fixture."""
    try:
        backend = importlib.import_module('cumin.backends.{backend}'.format(backend=backend_name))
    except ImportError:
        return  # Backend not available

    results = backend.grammar().runTests(
        get_fixture(os.path.join(BASE_PATH, '{backend}_invalid.txt'.format(backend=backend_name)), as_string=True),
        failureTests=True)
    assert results[0]



"""Abstract query tests."""

import pytest

from cumin.backends import BaseQuery


def test_base_query_instantiation():
    """Class BaseQuery is not instantiable being an abstract class."""
    with pytest.raises(TypeError):
        BaseQuery({})  # pylint: disable=abstract-class-instantiated



"""Transport ClusterShell: worker and event handlers."""
import logging
import sys
import threading

from collections import Counter, defaultdict

from ClusterShell import Event, Task
from tqdm import tqdm

from cumin import nodeset, nodeset_fromlist
from cumin.color import Colored
from cumin.transports import BaseWorker, NoProgress, ProgressBars, raise_error, State, WorkerError


class ClusterShellWorker(BaseWorker):
    """It provides a Cumin worker for SSH using the ClusterShell library.

    This transport uses the :py:mod:`ClusterShell` Python library to connect to the selected hosts and execute a
    list of commands. This transport accept the following customizations:

    * ``sync`` execution mode: given a list of commands, the first one will be executed on all the hosts, then, if the
      success ratio is reached, the second one will be executed on all hosts where the first one was successful, and so
      on.
    * ``async`` execution mode: given a list of commands, on each hosts the commands will be executed sequentially,
      interrupting the execution on any single host at the first command that fails. The execution on the hosts is
      independent between each other.
    * custom execution mode: can be achieved creating a custom event handler class that extends the ``BaseEventHandler``
      class defined in ``cumin/transports/clustershell.py``, implementing its abstract methods and setting to this class
      object the handler to the transport.
    """

    def __init__(self, config, target):
        """Worker ClusterShell constructor.

        :Parameters:
            according to parent :py:meth:`cumin.transports.BaseWorker.__init__`.
        """
        super().__init__(config, target)
        self.task = Task.task_self()  # Initialize a ClusterShell task
        self._handler_instance = None

        # Set any ClusterShell task options
        for key, value in config.get('clustershell', {}).items():
            if isinstance(value, list):
                self.task.set_info(key, ' '.join(value))
            else:
                self.task.set_info(key, value)

    def execute(self):
        """Execute the commands on all the targets using the handler.

        Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.transports.BaseWorker.execute`.
        """
        if not self.commands:
            raise WorkerError('No commands provided.')

        if self.handler is None:
            raise WorkerError('An EventHandler is mandatory.')

        # Instantiate handler
        # Schedule only the first command for the first batch, the following ones must be handled by the EventHandler
        self._handler_instance = self.handler(  # pylint: disable=not-callable
            self.target, self.commands, success_threshold=self.success_threshold)

        self.logger.info(
            "Executing commands %s on '%d' hosts: %s", self.commands, len(self.target.hosts), self.target.hosts)
        self.task.shell(self.commands[0].command, nodes=self.target.first_batch, handler=self._handler_instance,
                        timeout=self.commands[0].timeout, stdin=False)

        return_value = 0
        try:
            self.task.run(timeout=self.timeout, stdin=False)
            self.task.join()
        except Task.TimeoutError:
            if self._handler_instance is not None:
                self._handler_instance.on_timeout(self.task)
        finally:
            if self._handler_instance is not None:
                self._handler_instance.close(self.task)
                return_value = self._handler_instance.return_value
                if return_value is None:
                    return_value = 3  # The handler did not set a return value

        return return_value

    def get_results(self):
        """Get the results of the last task execution.

        Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.transports.BaseWorker.get_results`.
        """
        for output, nodelist in self.task.iter_buffers():
            yield nodeset_fromlist(nodelist), output

    @property
    def handler(self):
        """Concrete implementation of parent abstract getter and setter.

        Accepted values for the setter:
        * an instance of a custom handler class derived from :py:class:`BaseEventHandler`.
        * a :py:class:`str` with one of the available default handler listed in :py:data:`DEFAULT_HANDLERS`.

        The event handler is mandatory for this transport.

        :Parameters:
            according to parent :py:attr:`cumin.transports.BaseWorker.handler`.
        """
        return self._handler

    @handler.setter
    def handler(self, value):
        """Setter for the `handler` property. The relative documentation is in the getter."""
        if isinstance(value, type) and issubclass(value, BaseEventHandler):
            self._handler = value
        elif value in DEFAULT_HANDLERS:
            self._handler = DEFAULT_HANDLERS[value]
        else:
            raise_error(
                'handler',
                'must be one of ({default}, a class object derived from BaseEventHandler)'.format(
                    default=', '.join(DEFAULT_HANDLERS.keys())),
                value)


class Node:
    """Node class to represent each target node.

    Additional attributes available are:

    .. attribute:: state

       :py:class:`cumin.transports.State`: the state of the node.

    .. attribute:: running_command_index

       :py:class:`int`: the index of the current running command in the list of commands.

    """

    def __init__(self, name, commands):
        """Node class constructor with default values.

        Arguments:
            name (str): the hostname of the node.
            commands (list): a list of :py:class:`cumin.transports.Command` objects to be executed on the node.

        """
        self.name = name
        self.commands = commands
        self.state = State()  # Initialize the state machine for this node.
        self.running_command_index = -1  # Pointer to the current running command in self.commands


class BaseEventHandler(Event.EventHandler):
    """ClusterShell event handler base class.

    Inherit from :py:class:`ClusterShell.Event.EventHandler` class and define a base `EventHandler` class to be used
    in Cumin. It can be subclassed to generate custom `EventHandler` classes while taking advantage of some common
    functionalities.
    """

    short_command_length = 35
    """:py:class:`int`: the length to which a command should be shortened in various outputs."""

    def __init__(self, target, commands, success_threshold=1.0, progress_bars=True, **kwargs):
        """Event handler ClusterShell extension constructor.

        Arguments:
            target (cumin.transports.Target): a Target instance.
            commands (list): the list of Command objects that has to be executed on the nodes.
            success_threshold (float, optional): the success threshold, a :py:class:`float` between ``0`` and ``1``,
                to consider the execution successful.
            progress_bars (bool): should progress bars be displayed
            **kwargs (optional): additional keyword arguments that might be used by derived classes.

        """
        super().__init__()
        self.success_threshold = success_threshold
        self.logger = logging.getLogger('.'.join((self.__module__, self.__class__.__name__)))
        self.target = target
        self.lock = threading.Lock()  # Used to update instance variables coherently from within callbacks

        # Execution management variables
        self.return_value = None
        self.commands = commands
        self.kwargs = kwargs  # Allow to store custom parameters from subclasses without changing the signature
        self.counters = Counter()
        self.counters['total'] = len(target.hosts)
        self.deduplicate_output = self.counters['total'] > 1
        self.global_timedout = False
        # Instantiate all the node instances, slicing the commands list to get a copy
        self.nodes = {node: Node(node, commands[:]) for node in target.hosts}
        # Move already all the nodes in the first_batch to the scheduled state, it means that ClusterShell was
        # already instructed to execute a command on those nodes
        for node_name in target.first_batch:
            self.nodes[node_name].state.update(State.scheduled)

        # Initialize progress bar formats
        if progress_bars:
            self.progress = ProgressBars()
        else:
            self.progress = NoProgress()

    def close(self, task):
        """Additional method called at the end of the whole execution, useful for reporting and final actions.

        Arguments:
            task (ClusterShell.Task.Task): a ClusterShell Task instance.

        """
        raise NotImplementedError

    def on_timeout(self, task):
        """Update the state of the nodes and the timeout counter.

        Callback called by the :py:class:`ClusterShellWorker` when a :py:exc:`ClusterShell.Task.TimeoutError` is
        raised. It means that the whole execution timed out.

        Arguments:
            task (ClusterShell.Task.Task): a ClusterShell Task instance.

        """
        num_timeout = task.num_timeout()
        self.logger.error('Global timeout was triggered while %d nodes were executing a command', num_timeout)

        with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
            self.global_timedout = True
            # Considering timed out also the nodes that were pending the execution (for example when executing in
            # batches) and those that were already scheduled (for example when the # of nodes is greater than
            # ClusterShell fanout)
            pending_or_scheduled = sum(
                (node.state.is_pending or node.state.is_scheduled
                 or (node.state.is_success and node.running_command_index < (len(node.commands) - 1))
                 ) for node in self.nodes.values())
            if pending_or_scheduled > 0:
                self.progress.update_failed(num_timeout + pending_or_scheduled)

        self._global_timeout_nodes_report()

    def ev_pickup(self, worker, node):
        """Command execution started on a node, remove the command from the node's queue.

        This callback is triggered by the `ClusterShell` library for each node when it starts executing a command.

        :Parameters:
            according to parent :py:meth:`ClusterShell.Event.EventHandler.ev_pickup`.
        """
        self.logger.debug("node=%s, command='%s'", node, worker.command)

        with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
            curr_node = self.nodes[node]
            curr_node.state.update(State.running)  # Update the node's state to running

            command = curr_node.commands[curr_node.running_command_index + 1].command
            # Security check, it should never be triggered
            if command != worker.command:
                raise WorkerError("ev_pickup: got unexpected command '{command}', expected '{expected}'".format(
                    command=command, expected=worker.command))
            curr_node.running_command_index += 1  # Move the pointer of the current command

        if not self.deduplicate_output:
            output_message = "----- OUTPUT of '{command}' -----".format(
                command=self._get_short_command(worker.command))
            tqdm.write(Colored.blue(output_message), file=sys.stdout)

    def ev_read(self, worker, node, _, msg):
        """Worker has data to read from a specific node. Print it if running on a single host.

        This callback is triggered by ClusterShell for each node when output is available.

        :Parameters:
            according to parent :py:meth:`ClusterShell.Event.EventHandler.ev_read`.
        """
        if self.deduplicate_output:
            return

        with self.lock:
            tqdm.write(msg.decode(), file=sys.stdout)

    def ev_close(self, worker, timedout):
        """Worker has finished or timed out.

        This callback is triggered by ClusterShell when the execution has completed or timed out.

        :Parameters:
            according to parent :py:meth:`ClusterShell.Event.EventHandler.ev_close`.
        """
        if not timedout:
            return

        delta_timeout = worker.task.num_timeout() - self.counters['timeout']
        self.logger.debug("command='%s', delta_timeout=%d", worker.command, delta_timeout)

        with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
            self.progress.update_failed(delta_timeout)
            self.counters['timeout'] = worker.task.num_timeout()
            for node in worker.task.iter_keys_timeout():
                if not self.nodes[node].state.is_timeout:
                    self.nodes[node].state.update(State.timeout)

        # Schedule a timer to run the current command on the next node or start the next command
        worker.task.timer(self.target.batch_sleep, worker.eh)

    def _get_log_message(self, num, message, nodes=None):
        """Get a pre-formatted message suitable for logging or printing.

        Arguments:
            num (int): the number of affecte nodes.
            message (str): the message to print.
            nodes (list, optional): the list of nodes affected.

        Returns:
            tuple: a tuple of ``(logging message, NodeSet of the affected nodes)``.

        """
        if nodes is None:
            nodes_string = ''
            message_end = ''
        else:
            nodes_string = nodeset_fromlist(nodes)
            message_end = ': '

        tot = self.counters['total']
        log_message = '{perc:.1%} ({num}/{tot}) {message}{message_end}'.format(
            perc=(num / tot), num=num, tot=tot, message=message, message_end=message_end)

        return (log_message, str(nodes_string))

    def _print_report_line(self, message, color_func=Colored.red, nodes_string=''):  # pylint: disable=no-self-use
        """Print a tqdm-friendly colored status line with success/failure ratio and optional list of nodes.

        Arguments:
            message (str): the message to print.
            color_func (function, optional): the coloring function, one of :py:class`cumin.color.Colored` methods.
            nodes_string (str, optional): the string representation of the affected nodes.

        """
        tqdm.write(color_func(message) + Colored.cyan(nodes_string), file=sys.stderr)

    def _get_short_command(self, command):
        """Return a shortened representation of a command omitting the central part, if it's too long.

        Arguments:
            command (str): the command to be shortened.

        Returns:
            str: the short command.

        """
        sublen = (self.short_command_length - 3) // 2  # The -3 is for the ellipsis
        return (command[:sublen] + '...' + command[-sublen:]) if len(command) > self.short_command_length else command

    def _commands_output_report(self, buffer_iterator, command=None):
        """Print the commands output in a colored and tqdm-friendly way.

        Arguments:
            buffer_iterator (mixed): any `ClusterShell` object that implements ``iter_buffers()`` like
                :py:class:`ClusterShell.Task.Task` and all the `Worker` objects.
            command (str, optional): the command the output is referring to.

        """
        if not self.deduplicate_output:
            tqdm.write(Colored.blue('================'), file=sys.stdout)
            return

        nodelist = None
        if command is not None:
            output_message = "----- OUTPUT of '{command}' -----".format(command=self._get_short_command(command))
        else:
            output_message = '----- OUTPUT -----'

        for output, nodelist in buffer_iterator.iter_buffers():
            tqdm.write(Colored.blue('===== NODE GROUP ====='), file=sys.stdout)
            tqdm.write(Colored.cyan('({num}) {nodes}'.format(num=len(nodelist), nodes=nodeset_fromlist(nodelist))),
                       file=sys.stdout)
            tqdm.write(Colored.blue(output_message), file=sys.stdout)
            tqdm.write(output.message().decode(), file=sys.stdout)

        if nodelist is None:
            message = '===== NO OUTPUT ====='
        else:
            message = '================'

        tqdm.write(Colored.blue(message), file=sys.stdout)

    def _global_timeout_nodes_report(self):
        """Print the nodes that were caught by the global timeout in a colored and tqdm-friendly way."""
        if not self.global_timedout:
            return

        timeout = [node.name for node in self.nodes.values() if node.state.is_timeout]
        timeout_desc = 'of nodes were executing a command when the global timeout occurred'
        timeout_message, timeout_nodes = self._get_log_message(len(timeout), timeout_desc, nodes=timeout)
        self.logger.error('%s%s', timeout_message, timeout_nodes)
        self._print_report_line(timeout_message, nodes_string=timeout_nodes)

        not_run = [node.name for node in self.nodes.values() if node.state.is_pending or node.state.is_scheduled]
        not_run_desc = 'of nodes were pending execution when the global timeout occurred'
        not_run_message, not_run_nodes = self._get_log_message(len(not_run), not_run_desc, nodes=not_run)
        self.logger.error('%s%s', not_run_message, not_run_nodes)
        self._print_report_line(not_run_message, nodes_string=not_run_nodes)

    def _failed_commands_report(self, filter_command_index=-1):
        """Print the nodes that failed to execute commands in a colored and tqdm-friendly way.

        Arguments:
            filter_command_index (int, optional): print only the nodes that failed to execute the command specified by
                this command index.

        """
        for state in (State.failed, State.timeout):
            failed_commands = defaultdict(list)
            for node in [node for node in self.nodes.values() if node.state == state]:
                failed_commands[node.running_command_index].append(node.name)

            for index, nodes in failed_commands.items():
                command = self.commands[index].command

                if filter_command_index >= 0 and command is not None and index != filter_command_index:
                    continue

                message = "of nodes {state} to execute command '{command}'".format(
                    state=State.states_representation[state], command=self._get_short_command(command))
                log_message, nodes_string = self._get_log_message(len(nodes), message, nodes=nodes)
                self.logger.error('%s%s', log_message, nodes_string)
                self._print_report_line(log_message, nodes_string=nodes_string)

    def _success_nodes_report(self, command=None):
        """Print how many nodes succesfully executed all commands in a colored and tqdm-friendly way.

        Arguments:
            command (str, optional): the command the report is referring to.

        """
        if self.global_timedout and command is None:
            num = sum(1 for node in self.nodes.values() if node.state.is_success
                      and node.running_command_index == (len(self.commands) - 1))
        else:
            num = self.counters['success']

        tot = self.counters['total']
        success_ratio = num / tot

        if success_ratio < self.success_threshold:
            comp = '<'
            post = '. Aborting.'
        else:
            comp = '>='
            post = '.'

        message_string = ' of nodes successfully executed all commands'
        if command is not None:
            message_string = " for command: '{command}'".format(command=self._get_short_command(command))

        nodes = None
        if num not in (0, tot):
            nodes = [node.name for node in self.nodes.values() if node.state.is_success]

        message = "success ratio ({comp} {perc:.1%} threshold){message_string}{post}".format(
            comp=comp, perc=self.success_threshold, message_string=message_string, post=post)
        log_message, nodes_string = self._get_log_message(num, message, nodes=nodes)

        if num == tot:
            color_func = Colored.green
            level = logging.INFO
        elif success_ratio >= self.success_threshold:
            color_func = Colored.yellow
            level = logging.WARNING
        else:
            color_func = Colored.red
            level = logging.CRITICAL

        self.logger.log(level, '%s%s', log_message, nodes_string)
        self._print_report_line(log_message, color_func=color_func, nodes_string=nodes_string)


class SyncEventHandler(BaseEventHandler):
    """Custom ClusterShell event handler class that execute commands synchronously.

    The implemented logic is:

    * execute command `#N` on all nodes where command #`N-1` was successful according to `batch_size`.
    * the success ratio is checked at each command completion on every node, and will abort if not met, however
      nodes already scheduled for execution with `ClusterShell` will execute the command anyway. The use of the
      `batch_size` allow to control this aspect.
    * if the execution of command `#N` is completed and the success ratio is greater than the success threshold,
      re-start from the top with `N=N+1`.

    The typical use case is to orchestrate some operation across a fleet, ensuring that each command is completed by
    enough nodes before proceeding with the next one.
    """

    def __init__(self, target, commands, success_threshold=1.0, progress_bars=True, **kwargs):
        """Define a custom ClusterShell event handler to execute commands synchronously.

        :Parameters:
            according to parent :py:meth:`BaseEventHandler.__init__`.

        """
        super().__init__(
            target, commands, success_threshold=success_threshold, progress_bars=progress_bars, **kwargs)
        self.current_command_index = 0  # Global pointer for the current command in execution across all nodes
        self.start_command()
        self.aborted = False

    def start_command(self, schedule=False):
        """Initialize progress bars and variables for this command execution.

        Executed at the start of each command.

        Arguments:
            schedule (bool, optional): whether the next command should be sent to ClusterShell for execution or not.

        """
        self.counters['success'] = 0

        self.progress.init(self.counters['total'])

        # Schedule the next command, the first was already scheduled by ClusterShellWorker.execute()
        if schedule:
            with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
                # Available nodes for the next command execution were already update back to the pending state
                remaining_nodes = [node.name for node in self.nodes.values() if node.state.is_pending]
                first_batch = remaining_nodes[:self.target.batch_size]
                first_batch_set = nodeset_fromlist(first_batch)
                for node_name in first_batch:
                    self.nodes[node_name].state.update(State.scheduled)

            command = self.commands[self.current_command_index]
            self.logger.debug(
                "command='%s', timeout=%s, first_batch=%s", command.command, command.timeout, first_batch_set)

            # Schedule the command for execution in ClusterShell
            Task.task_self().flush_buffers()
            Task.task_self().shell(command.command, nodes=first_batch_set, handler=self, timeout=command.timeout)

    def end_command(self):
        """Command terminated, print the result and schedule the next command if criteria are met.

        Executed at the end of each command inside a lock.

        Returns:
            bool: :py:data:`True` if the next command should be scheduled, :py:data:`False` otherwise.

        """
        self._commands_output_report(Task.task_self(), command=self.commands[self.current_command_index].command)

        self.progress.close()

        self._failed_commands_report(filter_command_index=self.current_command_index)
        self._success_nodes_report(command=self.commands[self.current_command_index].command)

        success_ratio = self.counters['success'] / self.counters['total']

        # Abort on failure
        if success_ratio < self.success_threshold:
            self.return_value = 2
            self.aborted = True  # Tells other timers that might trigger after that the abort is already in progress
            return False

        if success_ratio == 1:
            self.return_value = 0
        else:
            self.return_value = 1

        if self.current_command_index == (len(self.commands) - 1):
            self.logger.debug('This was the last command')
            return False  # This was the last command

        return True

    def on_timeout(self, task):
        """Override parent class `on_timeout` method to run `end_command`.

        :Parameters:
            according to parent :py:meth:`BaseEventHandler.on_timeout`.
        """
        super().on_timeout(task)
        self.end_command()

    def ev_hup(self, worker, node, rc):
        """Command execution completed on a node.

        This callback is triggered by ClusterShell for each node when it completes the execution of a command.
        Update the progress bars and keep track of nodes based on the success/failure of the command's execution.
        Schedule a timer for further decisions.

        :Parameters:
            according to parent :py:meth:`ClusterShell.Event.EventHandler.ev_hup`.
        """
        self.logger.debug("node=%s, rc=%d, command='%s'", node, rc, worker.command)

        with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
            curr_node = self.nodes[node]

            ok_codes = curr_node.commands[curr_node.running_command_index].ok_codes
            if rc in ok_codes or not ok_codes:
                self.progress.update_success()
                self.counters['success'] += 1
                new_state = State.success
            else:
                self.progress.update_failed()
                self.counters['failed'] += 1
                new_state = State.failed

            curr_node.state.update(new_state)

        # Schedule a timer to run the current command on the next node or start the next command
        worker.task.timer(self.target.batch_sleep, worker.eh)

    def ev_timer(self, timer):  # noqa, mccabe: MC0001 too complex (15) FIXME
        """Schedule the current command on the next node or the next command on the first batch of nodes.

        This callback is triggered by `ClusterShell` when a scheduled `Task.timer()` goes off.

        :Parameters:
            according to parent :py:meth:`ClusterShell.Event.EventHandler.ev_timer`.
        """
        success_ratio = 1 - ((self.counters['failed'] + self.counters['timeout']) / self.counters['total'])

        node = None
        if success_ratio >= self.success_threshold:
            # Success ratio is still good, looking for the next node
            with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
                for new_node in self.nodes.values():
                    if new_node.state.is_pending:
                        # Found the next node where to execute the command
                        node = new_node
                        node.state.update(State.scheduled)
                        break

        if node is not None:
            # Schedule the execution with ClusterShell of the current command to the next node found above
            command = self.nodes[node.name].commands[self.nodes[node.name].running_command_index + 1]
            self.logger.debug("next_node=%s, timeout=%s, command='%s'", node.name, command.command, command.timeout)
            Task.task_self().shell(command.command, handler=timer.eh, timeout=command.timeout, nodes=nodeset(node.name))
            return

        # No more nodes were left for the execution of the current command
        with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
            try:
                command = self.commands[self.current_command_index].command
            except IndexError:
                command = None  # Last command reached

            # Get a list of the nodes still in pending state
            pending = [pending_node.name for pending_node in self.nodes.values() if pending_node.state.is_pending]
            # Nodes in running are still running the command and nodes in scheduled state will execute the command
            # anyway, they were already offloaded to ClusterShell
            accounted = len(pending) + self.counters['failed'] + self.counters['success'] + self.counters['timeout']

            # Avoid race conditions
            if self.aborted or accounted != self.counters['total'] or command is None or self.global_timedout:
                self.logger.debug("Skipped timer")
                return

            if pending:
                # This usually happens when executing in batches
                self.logger.warning("Command '%s' was not executed on: %s", command, nodeset_fromlist(pending))

            self.logger.info("Completed command '%s'", command)
            restart = self.end_command()
            self.current_command_index += 1  # Move the global pointer of the command in execution

            if restart:
                for node in self.nodes.values():
                    if node.state.is_success:
                        # Only nodes in pending state will be scheduled for the next command
                        node.state.update(State.pending)

        if restart:
            self.start_command(schedule=True)

    def close(self, task):
        """Concrete implementation of parent abstract method to print the success nodes report.

        :Parameters:
            according to parent :py:meth:`cumin.transports.BaseEventHandler.close`.
        """
        self._success_nodes_report()


class AsyncEventHandler(BaseEventHandler):
    """Custom ClusterShell event handler class that execute commands asynchronously.

    The implemented logic is:

    * execute on all nodes independently every command in a sequence, aborting the execution on that node if any
      command fails.
    * The success ratio is checked at each node completion (either because it completed all commands or aborted
      earlier), however nodes already scheduled for execution with ClusterShell will execute the commands anyway. The
      use of the batch_size allows to control this aspect.
    * if the success ratio is met, schedule the execution of all commands to the next node.

    The typical use case is to execute read-only commands to gather the status of a fleet without any special need of
    orchestration between the nodes.
    """

    def __init__(self, target, commands, success_threshold=1.0, progress_bars=True, **kwargs):
        """Define a custom ClusterShell event handler to execute commands asynchronously between nodes.

        :Parameters:
            according to parent :py:meth:`BaseEventHandler.__init__`.
        """
        super().__init__(
            target, commands, success_threshold=success_threshold, progress_bars=progress_bars, **kwargs)

        self.progress.init(self.counters['total'])

    def ev_hup(self, worker, node, rc):
        """Command execution completed on a node.

        This callback is triggered by ClusterShell for each node when it completes the execution of a command.
        Enqueue the next command if the success criteria are met, track the failure otherwise. Update the progress
        bars accordingly.

        :Parameters:
            according to parent :py:meth:`ClusterShell.Event.EventHandler.ev_hup`.
        """
        self.logger.debug("node=%s, rc=%d, command='%s'", node, rc, worker.command)

        schedule_next = False
        schedule_timer = False
        with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
            curr_node = self.nodes[node]

            ok_codes = curr_node.commands[curr_node.running_command_index].ok_codes
            if rc in ok_codes or not ok_codes:
                if curr_node.running_command_index == (len(curr_node.commands) - 1):
                    self.progress.update_success()
                    self.counters['success'] += 1
                    curr_node.state.update(State.success)
                    schedule_timer = True  # Continue the execution on other nodes if criteria are met
                else:
                    schedule_next = True  # Continue the execution in the current node with the next command
            else:
                self.progress.update_failed()
                self.counters['failed'] += 1
                curr_node.state.update(State.failed)
                schedule_timer = True  # Continue the execution on other nodes if criteria are met

        if schedule_next:
            # Schedule the execution of the next command on this node with ClusterShell
            command = curr_node.commands[curr_node.running_command_index + 1]
            worker.task.shell(
                command.command, nodes=nodeset(node), handler=worker.eh, timeout=command.timeout, stdin=False)
        elif schedule_timer:
            # Schedule a timer to allow to run all the commands in the next available node
            worker.task.timer(self.target.batch_sleep, worker.eh)

    def ev_timer(self, timer):
        """Schedule the current command on the next node or the next command on the first batch of nodes.

        This callback is triggered by `ClusterShell` when a scheduled `Task.timer()` goes off.

        :Parameters:
            according to parent :py:meth:`ClusterShell.Event.EventHandler.ev_timer`.
        """
        success_ratio = 1 - ((self.counters['failed'] + self.counters['timeout']) / self.counters['total'])

        node = None
        if success_ratio >= self.success_threshold:
            # Success ratio is still good, looking for the next node
            with self.lock:  # Avoid modifications of the same data from other callbacks triggered by ClusterShell
                for new_node in self.nodes.values():
                    if new_node.state.is_pending:
                        # Found the next node where to execute all the commands
                        node = new_node
                        node.state.update(State.scheduled)
                        break

        if node is not None:
            # Schedule the exeuction of the first command to the next node with ClusterShell
            command = node.commands[0]
            self.logger.debug("next_node=%s, timeout=%s, command='%s'", node.name, command.command, command.timeout)
            Task.task_self().shell(
                command.command, handler=timer.eh, timeout=command.timeout, nodes=nodeset(node.name))
        else:
            self.logger.debug('No more nodes left')

    def close(self, task):
        """Concrete implementation of parent abstract method to print the nodes reports and close progress bars.

        :Parameters:
            according to parent :py:meth:`cumin.transports.BaseEventHandler.close`.
        """
        self._commands_output_report(task)

        self.progress.close()

        self._failed_commands_report()
        self._success_nodes_report()

        num = self.counters['success']
        tot = self.counters['total']
        success_ratio = num / tot

        if success_ratio == 1:
            self.return_value = 0
        elif success_ratio < self.success_threshold:
            self.return_value = 2
        else:
            self.return_value = 1


worker_class = ClusterShellWorker  # pylint: disable=invalid-name
"""Required by the transport auto-loader in :py:meth:`cumin.transport.Transport.new`."""

DEFAULT_HANDLERS = {'sync': SyncEventHandler, 'async': AsyncEventHandler}
"""dict: mapping of available default event handlers for :py:class:`ClusterShellWorker`."""



"""Abstract transport and state machine for hosts state."""
import logging
import os
import shlex
import sys

from abc import ABCMeta, abstractmethod, abstractproperty

from ClusterShell.NodeSet import NodeSet
from tqdm import tqdm

from cumin import CuminError, nodeset_fromlist
from cumin.color import Colored


class WorkerError(CuminError):
    """Custom exception class for worker errors."""


class StateTransitionError(CuminError):
    """Exception raised when an invalid transition for a node's State was attempted."""


class InvalidStateError(CuminError):
    """Exception raised when an invalid transition for a node's State was attempted."""


class Command:
    """Class to represent a command."""

    def __init__(self, command, timeout=None, ok_codes=None):
        """Command constructor.

        Arguments:
            command (str): the command to execute.
            timeout (int, optional): the command's timeout in seconds.
            ok_codes (list, optional): a list of exit codes to be considered successful for the command.
                The exit code zero is considered successful by default, if this option is set it override it. If set
                to an empty list ``[]``, it means that any code is considered successful.

        """
        self.command = command
        self._timeout = None
        self._ok_codes = None

        if timeout is not None:
            self.timeout = timeout

        if ok_codes is not None:
            self.ok_codes = ok_codes

    def __repr__(self):
        """Return the representation of the :py:class:`Command`.

        The representation allow to instantiate a new :py:class:`Command` instance with the same properties.

        Returns:
            str: the representation of the object.

        """
        params = ["'{command}'".format(command=self.command.replace("'", r'\''))]

        for field in ('_timeout', '_ok_codes'):
            value = getattr(self, field)
            if value is not None:
                params.append('{key}={value}'.format(key=field[1:], value=value))

        return 'cumin.transports.Command({params})'.format(params=', '.join(params))

    def __str__(self):
        """Return the string representation of the command.

        Returns:
            str: the string representation of the object.

        """
        return self.command

    def __eq__(self, other):
        """Equality operation. Allow to directly compare a :py:class:`Command` object to another or a string.

        :Parameters:
            according to Python's Data model :py:meth:`object.__eq__`.

        Returns:
            bool: :py:data:`True` if the `other` object is equal to this one, :py:data:`False` otherwise.

        Raises:
            exceptions.ValueError: if the comparing object is not an instance of :py:class:`Command` or a
                :py:class:`str`.

        """
        if isinstance(other, str):
            other_command = other
            same_params = (self._timeout is None and self._ok_codes is None)
        elif isinstance(other, Command):
            other_command = other.command
            same_params = (self.timeout == other.timeout and self.ok_codes == other.ok_codes)
        else:
            raise ValueError("Unable to compare instance of '{other}' with Command instance".format(other=type(other)))

        return shlex.split(self.command) == shlex.split(other_command) and same_params

    def __ne__(self, other):
        """Inequality operation. Allow to directly compare a Command object to another or a string.

        :Parameters:
            according to Python's Data model :py:meth:`object.__ne__`.

        Returns:
            bool: :py:data:`True` if the `other` object is different to this one, :py:data:`False` otherwise.

        Raises:
            exceptions.ValueError: if the comparing object is not an instance of :py:class:`Command` or a
                :py:class:`str`.

        """
        return not self == other

    @property
    def timeout(self):
        """Timeout of the :py:class:`Command`.

        :Getter:
            Returns the current `timeout` or :py:data:`None` if not set.

        :Setter:
            :py:class:`float`, :py:class:`int`, :py:data:`None`: the `timeout` in seconds for the execution of the
            `command` on each host. Both :py:class:`float` and :py:class:`int` are accepted and converted internally to
            :py:class:`float`. If :py:data:`None` the `timeout` is reset to its default value.

        Raises:
            cumin.transports.WorkerError: if trying to set it to an invalid value.

        """
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        """Setter for the timeout property. The relative documentation is in the getter."""
        if isinstance(value, int):
            value = float(value)

        validate_positive_float('timeout', value)
        self._timeout = value

    @property
    def ok_codes(self):
        """List of exit codes to be considered successful for the execution of the :py:class:`Command`.

        :Getter:
            Returns the current `ok_codes` or a :py:class:`list` with the element ``0`` if not set.

        :Setter:
            :py:class:`list[int]`, :py:data:`None`: list of exit codes to be considered successful for the execution of
            the `command` on each host. Must be a :py:class:`list` of :py:class:`int` in the range ``0-255`` included,
            or :py:data:`None` to unset it. The exit code ``0`` is considered successful by default, but it can be
            overriden setting this property. Set it to an empty :py:class:`list` to consider any
            exit code successful.

        Raises:
            cumin.transports.WorkerError: if trying to set it to an invalid value.

        """
        ok_codes = self._ok_codes
        if ok_codes is None:
            ok_codes = [0]

        return ok_codes

    @ok_codes.setter
    def ok_codes(self, value):
        """Setter for the ok_codes property. The relative documentation is in the getter."""
        if value is None:
            self._ok_codes = value
            return

        validate_list('ok_codes', value, allow_empty=True)
        for code in value:
            if not isinstance(code, int) or code < 0 or code > 255:
                raise_error('ok_codes', 'must be a list of integers in the range 0-255 or None', value)

        self._ok_codes = value


class State:
    """State machine for the state of a host.

    .. attribute:: current

       :py:class:`int`: the current state.

    .. attribute:: is_pending

       :py:class:`bool`: :py:data:`True` if the current state is `pending`, :py:data:`False` otherwise.

    .. attribute:: is_scheduled

       :py:class:`bool`: :py:data:`True` if the current state is `scheduled`, :py:data:`False` otherwise.

    .. attribute:: is_running

       :py:class:`bool`: :py:data:`True` if the current state is `running`, :py:data:`False` otherwise.

    .. attribute:: is_success

       :py:class:`bool`: :py:data:`True` if the current state is `success`, :py:data:`False` otherwise.

    .. attribute:: is_failed

       :py:class:`bool`: :py:data:`True` if the current state is `failed`, :py:data:`False` otherwise.

    .. attribute:: is_timeout

       :py:class:`bool`: :py:data:`True` if the current state is `timeout`, :py:data:`False` otherwise.

    """

    valid_states = range(6)
    """:py:class:`list`: valid states integer indexes."""

    pending, scheduled, running, success, failed, timeout = valid_states
    """Valid state property, one for each :py:data:`cumin.transports.State.valid_states`."""

    states_representation = ('pending', 'scheduled', 'running', 'success', 'failed', 'timeout')
    """:py:func:`tuple`: Tuple with the string representations of the valid states."""

    allowed_state_transitions = {
        pending: (scheduled, ),
        scheduled: (running, ),
        running: (running, success, failed, timeout),
        success: (pending, ),
        failed: (),
        timeout: (),
    }
    """:py:class:`dict`: Dictionary with ``{valid state: tuple of valid states}`` mapping of the allowed transitions
    between all the possile states.

    This is the diagram of the allowed transitions:

    .. image:: ../../examples/transports_state_transitions.png
       :alt: State class allowed transitions diagram
    |

    """

    def __init__(self, init=None):
        """State constructor. The initial state is set to `pending` it not provided.

        Arguments:
            init (int, optional): the initial state from where to start. The `pending` state will be used if not set.

        Raises:
            cumin.transports.InvalidStateError: if `init` is an invalid state.

        """
        if init is None:
            self._state = self.pending
        elif init in self.valid_states:
            self._state = init
        else:
            raise InvalidStateError("Initial state '{state}' is not a valid state. Expected one of {states}".format(
                state=init, states=self.valid_states))

    def __getattr__(self, name):
        """Attribute accessor.

        :Accessible properties:
            * `current` (:py:class:`int`): retuns the current state.
            * `is_{valid_state_name}` (:py:class:`bool`): for each valid state name, returns :py:data:`True` if the
              current state matches the state in the variable name. :py:data:`False` otherwise.

        :Parameters:
            according to Python's Data model :py:meth:`object.__getattr__`.

        Raises:
            exceptions.AttributeError: if the attribute name is not available.

        """
        if name == 'current':
            return self._state

        if name.startswith('is_') and name[3:] in self.states_representation:
            return getattr(self, name[3:]) == self._state

        raise AttributeError("'State' object has no attribute '{name}'".format(name=name))

    def __repr__(self):
        """Return the representation of the :py:class:`State`.

        The representation allow to instantiate a new :py:class:`State` instance with the same properties.

        Returns:
            str: the representation of the object.

        """
        return 'cumin.transports.State(init={state})'.format(state=self._state)

    def __str__(self):
        """Return the string representation of the state.

        Returns:
            str: the string representation of the object.

        """
        return self.states_representation[self._state]

    def __eq__(self, other):
        """Equality operator for rich comparison.

        :Parameters:
            according to Python's Data model :py:meth:`object.__eq__`.

        Returns:
            bool: :py:data:`True` if `self` is equal to `other`, :py:data:`False` otherwise.

        Raises:
            exceptions.ValueError: if the comparing object is not an instance of :py:class:`State` or a
                :py:class:`int`.

        """
        return self._cmp(other) == 0

    def __lt__(self, other):
        """Less than operator for rich comparison.

        :Parameters:
            according to Python's Data model :py:meth:`object.__lt__`.

        Returns:
            bool: :py:data:`True` if `self` is lower than `other`, :py:data:`False` otherwise.

        Raises:
            exceptions.ValueError: if the comparing object is not an instance of :py:class:`State` or a
                :py:class:`int`.

        """
        return self._cmp(other) < 0

    def __le__(self, other):
        """Less than or equal operator for rich comparison.

        :Parameters:
            according to Python's Data model :py:meth:`object.__le__`.

        Returns:
            bool: :py:data:`True` if `self` is lower or equal than `other`, :py:data:`False` otherwise.

        Raises:
            exceptions.ValueError: if the comparing object is not an instance of :py:class:`State` or a
                :py:class:`int`.

        """
        return self._cmp(other) <= 0

    def __gt__(self, other):
        """Greater than operator for rich comparison.

        :Parameters:
            according to Python's Data model :py:meth:`object.__gt__`.

        Returns:
            bool: :py:data:`True` if `self` is greater than `other`, :py:data:`False` otherwise.

        Raises:
            exceptions.ValueError: if the comparing object is not an instance of :py:class:`State` or a
                :py:class:`int`.

        """
        return self._cmp(other) > 0

    def __ge__(self, other):
        """Greater than or equal operator for rich comparison.

        :Parameters:
            according to Python's Data model :py:meth:`object.__ge__`.

        Returns:
            bool: :py:data:`True` if `self` is greater or equal than `other`, :py:data:`False` otherwise.

        Raises:
            exceptions.ValueError: if the comparing object is not an instance of :py:class:`State` or a
                :py:class:`int`.

        """
        return self._cmp(other) >= 0

    def update(self, new):
        """Transition the state from the current state to the new one, if the transition is allowed.

        Arguments:
            new (int): the new state to set. Only specific state transitions are allowed.

        Raises:
            cumin.transports.StateTransitionError: if the transition is not allowed, see
                :py:attr:`allowed_state_transitions`.

        """
        if new not in self.valid_states:
            raise ValueError("State must be one of {valid}, got '{new}'".format(valid=self.valid_states, new=new))

        if new not in self.allowed_state_transitions[self._state]:
            raise StateTransitionError(
                "From the current state '{current}' the allowed states are '{allowed}', got '{new}'".format(
                    current=self._state, allowed=self.allowed_state_transitions[self._state], new=new))

        self._state = new

    def _cmp(self, other):
        """Comparison operation. Allow to directly compare a state object to another or to an integer.

        Arguments:
            other (mixed): the object to compare the current instance to.

        Raises:
            ValueError: if the comparing object is not an instance of State or an integer.

        """
        if isinstance(other, int):
            return self._state - other

        if isinstance(other, State):
            return self._state - other._state  # pylint: disable=protected-access

        raise ValueError("Unable to compare instance of '{other}' with State instance".format(other=type(other)))


class Target:
    """Targets management class."""

    def __init__(self, hosts, batch_size=None, batch_size_ratio=None, batch_sleep=None):
        """Constructor, inizialize the Target with the list of hosts and additional parameters.

        Arguments:
            hosts (ClusterShell.NodeSet.NodeSet, list): hosts that will be targeted, both
                :py:class:`ClusterShell.NodeSet.NodeSet` and :py:class:`list` are accepted and converted automatically
                to :py:class:`ClusterShell.NodeSet.NodeSet` internally.
            batch_size (int, optional): set the batch size so that no more that this number of hosts are targeted
                at any given time. It must be a positive integer. If greater than the number of hosts it will be
                auto-resized to the number of hosts.
            batch_size_ratio (float, optional): set the batch size with a ratio so that no more that this fraction
                of hosts are targeted at any given time. It must be a float between 0 and 1 and will raise exception
                if after rounding it there are 0 hosts selected.
            batch_sleep (float, optional): sleep time in seconds between the end of execution of one host in the
                batch and the start in the next host. It must be a positive float.

        Raises:
            cumin.transports.WorkerError: if the `hosts` parameter is empty or invalid, if both the `batch_size` and
                `batch_size_ratio` parameters are set or if the `batch_size_ratio` selects no hosts.

        """
        self.logger = logging.getLogger('.'.join((self.__module__, self.__class__.__name__)))

        message = "must be a non-empty ClusterShell NodeSet or list"
        if not hosts:
            raise_error('hosts', message, hosts)
        elif isinstance(hosts, NodeSet):
            self.hosts = hosts
        elif isinstance(hosts, list):
            self.hosts = nodeset_fromlist(hosts)
        else:
            raise_error('hosts', message, hosts)

        if batch_size is not None and batch_size_ratio is not None:
            raise WorkerError(("The 'batch_size' and 'batch_size_ratio' parameters are mutually exclusive but they're "
                               "both set."))

        if batch_size_ratio is not None:
            if not isinstance(batch_size_ratio, float) or not 0.0 <= batch_size_ratio <= 1.0:
                raise_error('batch_size_ratio', 'must be a float between 0.0 and 1.0', batch_size_ratio)

            batch_size = round(len(self.hosts) * batch_size_ratio)
            if batch_size == 0:
                raise_error('batch_size_ratio', 'has generated a batch_size of 0 hosts', batch_size_ratio)

        self.batch_size = self._compute_batch_size(batch_size, self.hosts)
        self.batch_sleep = Target._compute_batch_sleep(batch_sleep)

    @property
    def first_batch(self):
        """First batch of the hosts to target.

        :Getter:
            Returns a :py:class:`ClusterShell.NodeSet.NodeSet` of the first batch of hosts, according to the
            `batch_size`.
        """
        return self.hosts[:self.batch_size]

    def _compute_batch_size(self, batch_size, hosts):
        """Compute the batch_size based on the hosts size and return the value to be used.

        Arguments:
            batch_size (int, None): a positive integer to indicate the batch_size to apply when executing the worker or
                :py:data:`None` to get its default value of all the hosts. If greater than the number of hosts, the
                number of hosts will be used as value instead.
            hosts (ClusterShell.NodeSet.NodeSet): the list of hosts to use to calculate the batch size.

        Returns:
            int: the effective `batch_size` to use.

        """
        validate_positive_integer('batch_size', batch_size)
        hosts_size = len(hosts)

        if batch_size is None:
            batch_size = hosts_size
        elif batch_size > hosts_size:
            self.logger.debug(("Provided batch_size '%d' is greater than the number of hosts '%d'"
                               ", using '%d' as value"), batch_size, hosts_size, hosts_size)
            batch_size = hosts_size

        return batch_size

    @staticmethod
    def _compute_batch_sleep(batch_sleep):
        """Validate batch_sleep and return its value or a default value.

        Arguments:
            batch_sleep(float, None): a positive float indicating the sleep in seconds to apply between one batched
                host and the next, or :py:data:`None` to get its default value.

        Returns:
            float: the effective `batch_sleep` to use.

        """
        validate_positive_float('batch_sleep', batch_sleep)
        return batch_sleep or 0.0


class BaseWorker(metaclass=ABCMeta):
    """Worker interface to be extended by concrete workers."""

    def __init__(self, config, target):
        """Worker constructor. Setup environment variables and initialize properties.

        Arguments:
            config (dict): a dictionary with the parsed configuration file.
            target (Target): a Target instance.

        """
        self.config = config
        self.target = target
        self.logger = logging.getLogger('.'.join((self.__module__, self.__class__.__name__)))
        self.logger.trace('Transport %s created with config: %s', type(self).__name__, config)

        # Initialize setters values
        self._commands = None
        self._handler = None
        self._timeout = None
        self._success_threshold = None

        for key, value in config.get('environment', {}).items():
            os.environ[key] = value

    @abstractmethod
    def execute(self):
        """Execute the task as configured.

        Returns:
            int: ``0`` on success, a positive integer on failure.

        Raises:
            cumin.transports.WorkerError: if misconfigured.

        """

    @abstractmethod
    def get_results(self):
        """Iterate over the results (`generator`).

        Yields:
            tuple: with ``(hosts, result)`` for each host(s) of the current execution.

        """

    @property
    def commands(self):
        """Commands for the current execution.

        :Getter:
            Returns the current `command` :py:class:`list` or an empty :py:class:`list` if not set.

        :Setter:
            :py:class:`list[Command]`, :py:class:`list[str]`: a :py:class:`list` of :py:class:`Command` objects or
            :py:class:`str` to be executed in the hosts. The elements are converted to :py:class:`Command`
            automatically.

        Raises:
            cumin.transports.WorkerError: if trying to set it with invalid data.

        """
        return self._commands or []

    @commands.setter
    def commands(self, value):
        """Setter for the `commands` property. The relative documentation is in the getter."""
        if value is None:
            self._commands = value
            return

        validate_list('commands', value, allow_empty=True)
        commands = []
        for command in value:
            if isinstance(command, Command):
                commands.append(command)
            elif isinstance(command, str):
                commands.append(Command(command))
            else:
                raise_error('commands', 'must be a list of Command objects or strings', value)

        self._commands = commands

    @abstractproperty
    @property
    def handler(self):
        """Get and set the `handler` for the current execution.

        :Getter:
            Returns the current `handler` or :py:data:`None` if not set.

        :Setter:
            :py:class:`str`, :py:class:`EventHandler`, :py:data:`None`: an event handler to be notified of the progress
            during execution. Its interface depends on the actual transport chosen. Accepted values are:
            * None => don't use an event handler (default)
            * str => a string label to choose one of the available default EventHandler classes in that transport,
            * an event handler class object (not instance)
        """

    @abstractproperty
    @handler.setter
    def handler(self, value):  # pylint: disable=property-with-parameters; https://github.com/PyCQA/pylint/issues/3600
        """Setter for the `handler` property. The relative documentation is in the getter."""

    @property
    def timeout(self):
        """Global timeout for the current execution.

        :Getter:
            int: returns the current `timeout` or ``0`` (no timeout) if not set.

        :Setter:
            :py:class:`int`, :py:data:`None`: timeout for the current execution in seconds. Must be a positive integer
            or :py:data:`None` to reset it.

        Raises:
            cumin.transports.WorkerError: if trying to set it to an invalid value.

        """
        return self._timeout or 0

    @timeout.setter
    def timeout(self, value):
        """Setter for the global `timeout` property. The relative documentation is in the getter."""
        validate_positive_integer('timeout', value)
        self._timeout = value

    @property
    def success_threshold(self):
        """Success threshold for the current execution.

        :Getter:
            float: returns the current `success_threshold` or ``1.0`` (`100%`) if not set.

        :Setter:
            :py:class:`float`, :py:data:`None`: The success ratio threshold that must be reached to consider the run
            successful. A :py:class:`float` between ``0`` and ``1`` or :py:data:`None` to reset it. The specific
            meaning might change based on the chosen transport.

        Raises:
            cumin.transports.WorkerError: if trying to set it to an invalid value.

        """
        success_threshold = self._success_threshold
        if success_threshold is None:
            success_threshold = 1.0

        return success_threshold

    @success_threshold.setter
    def success_threshold(self, value):
        """Setter for the `success_threshold` property. The relative documentation is in the getter."""
        if value is not None and (not isinstance(value, float)
                                  or not (0.0 <= value <= 1.0)):  # pylint: disable=superfluous-parens
            raise WorkerError("success_threshold must be a float beween 0 and 1, got '{value_type}': {value}".format(
                value_type=type(value), value=value))

        self._success_threshold = value


def validate_list(property_name, value, allow_empty=False):
    """Validate a list.

    Arguments:
        property_name (str): the name of the property to validate.
        value (list): the value to validate.
        allow_empty (bool, optional): whether to consider an empty list valid.

    Raises:
        cumin.transports.WorkerError: if trying to set it to an invalid value.

    """
    if not isinstance(value, list):
        raise_error(property_name, 'must be a list', value)

    if not allow_empty and not value:
        raise_error(property_name, 'must be a non-empty list', value)


def validate_positive_integer(property_name, value):
    """Validate a positive integer or :py:data:`None`.

    Arguments:
        property_name (str): the name of the property to validate.
        value (int, None): the value to validate.

    Raises:
        cumin.transports.WorkerError: if trying to set it to an invalid value.

    """
    if value is not None and (not isinstance(value, int) or value <= 0):
        raise_error(property_name, 'must be a positive integer or None', value)


def validate_positive_float(property_name, value):
    """Validate a positive float or :py:data:`None`.

    Arguments:
        property_name (str): the name of the property to validate.
        value (float, None): the value to validate.

    Raises:
        cumin.transports.WorkerError: if trying to set it to an invalid value.

    """
    if value is not None and (not isinstance(value, float) or value <= 0):
        raise_error(property_name, 'must be a positive float or None', value)


def raise_error(property_name, message, value):
    """Raise a :py:class:`WorkerError` exception.

    Arguments:
        property_name (str): the name of the property that raised the exception.
        message (str): the message to use for the exception.
        value (mixed): the value that raised the exception.

    """
    raise WorkerError("{property_name} {message}, got '{value_type}': {value}".format(
        property_name=property_name, message=message, value_type=type(value), value=value))


class ProgressBars:
    """Progress bars for the status of successful / failed hosts.

    The ProgressBars needs to be notified of the total number of hosts when the
    operation starts, and then notified of successes and failures.
    """

    def __init__(self):
        """Create the progress bars.

        Note:
            the progress bars themselves are not initalized at object creation. ``init()`` needs to be called before
            using the progress bars.

        """
        self.pbar_ok = None
        self.pbar_ko = None
        self.bar_format = ('{desc} |{bar}| {percentage:3.0f}% ({n_fmt}/{total_fmt}) '
                           '[{elapsed}<{remaining}, {rate_fmt}]')

    def init(self, num_hosts):
        """Initialize the progress bars.

        Arguments:
            num_hosts (int): the total number of hosts

        """
        self.pbar_ok = tqdm(desc='PASS', total=num_hosts, leave=True, unit='hosts', dynamic_ncols=True,
                            bar_format=Colored.green(self.bar_format), file=sys.stderr)
        self.pbar_ok.refresh()
        self.pbar_ko = tqdm(desc='FAIL', total=num_hosts, leave=True, unit='hosts', dynamic_ncols=True,
                            bar_format=Colored.red(self.bar_format), file=sys.stderr)
        self.pbar_ko.refresh()

    def close(self):
        """Closes the progress bars."""
        self.pbar_ok.close()
        self.pbar_ko.close()

    def update_success(self, num_hosts=1):
        """Updates the number of successful hosts.

        Arguments:
            num_hosts (int): increment to the number of hosts that have completed successfully

        """
        self.pbar_ok.update(num_hosts)

    def update_failed(self, num_hosts=1):
        """Updates the number of failed hosts.

        Arguments:
            num_hosts (int): increment to the number of hosts that have completed in error

        """
        self.pbar_ko.update(num_hosts)


class NoProgress:
    """Used as a null object to disable the display of progress bars."""

    def init(self, num_hosts):
        """Does nothing."""

    def close(self):
        """Does nothing."""

    def update_success(self, num_hosts=1):
        """Does nothing."""

    def update_failed(self, num_hosts=1):
        """Does nothing."""



"""Color tests."""
from unittest import mock

import pytest

from cumin.color import Colored


def test_red():
    """It should return the message enclosed in ASCII red color code."""
    assert Colored.red('message') == '\x1b[31mmessage\x1b[39m'


def test_green():
    """It should return the message enclosed in ASCII green color code."""
    assert Colored.green('message') == '\x1b[32mmessage\x1b[39m'


def test_yellow():
    """It should return the message enclosed in ASCII yellow color code."""
    assert Colored.yellow('message') == '\x1b[33mmessage\x1b[39m'


def test_blue():
    """It should return the message enclosed in ASCII blue color code."""
    assert Colored.blue('message') == '\x1b[34mmessage\x1b[39m'


def test_cyan():
    """It should return the message enclosed in ASCII cyan color code."""
    assert Colored.cyan('message') == '\x1b[36mmessage\x1b[39m'


def test_wrong_case():
    """It should raise AttributeError if called with the wrong case."""
    with pytest.raises(AttributeError, match="'Colored' object has no attribute 'Red'"):
        Colored.Red('')


def test_non_existent():
    """It should raise AttributeError if called with a non existent color."""
    with pytest.raises(AttributeError, match="'Colored' object has no attribute 'missing'"):
        Colored.missing('')


def test_emtpy():
    """It should return an empty string if the object is empty."""
    assert Colored.red('') == ''


@mock.patch('cumin.color.Colored.disabled', new_callable=mock.PropertyMock)
def test_disabled(mocked_colored_disabled):
    """It should return the message untouched if coloration is disabled."""
    mocked_colored_disabled.return_value = True
    assert Colored.red('message') == 'message'
    assert mocked_colored_disabled.called



"""Pytest customization for unit tests."""
import re

import pytest
import requests_mock

from cumin.backends import puppetdb


def _requests_matcher_non_existent(request):
    return request.json() == {'query': '["or", ["=", "certname", "non_existent_host"]]'}


def _requests_matcher_invalid(request):
    return request.json() == {'query': '["or", ["=", "certname", "invalid_query"]]'}


@pytest.fixture()
def mocked_requests():
    """Set mocked requests fixture."""
    with requests_mock.Mocker() as mocker:
        yield mocker


@pytest.fixture(params=(3, 4))
def query_requests(request, mocked_requests):  # pylint: disable=redefined-outer-name
    """Set the requests library mock for each test and PuppetDB API version."""
    if request.param == 3:  # PuppetDB API v3
        query = puppetdb.PuppetDBQuery(
            {'puppetdb': {'api_version': 3, 'urllib3_disable_warnings': ['SubjectAltNameWarning']}})
        for endpoint, key in query.hosts_keys.items():
            mocked_requests.register_uri(
                'GET',
                re.compile(re.escape(query.url + endpoint + '?query=')),
                status_code=200,
                json=[{key: endpoint + '_host1', 'key': 'value1'}, {key: endpoint + '_host2', 'key': 'value2'}])

        # Register a requests response for a non matching query
        mocked_requests.register_uri(
            'GET', query.url + query.endpoints['F'] + '?query=["or", ["=", "name", "non_existent_host"]]',
            status_code=200, json=[], complete_qs=True)
        # Register a requests response for an invalid query
        mocked_requests.register_uri(
            'GET', query.url + query.endpoints['F'] + '?query=["or", ["=", "name", "invalid_query"]]',
            status_code=400, complete_qs=True)

    elif request.param == 4:  # PuppetDB API v4
        query = puppetdb.PuppetDBQuery({})
        for endpoint, key in query.hosts_keys.items():
            mocked_requests.register_uri(
                'POST', query.url + endpoint, status_code=200, complete_qs=True,
                json=[{key: endpoint + '_host1', 'key': 'value1'}, {key: endpoint + '_host2', 'key': 'value2'}])

        # Register a requests response for a non matching query
        mocked_requests.register_uri(
            'POST', query.url + query.endpoints['F'], status_code=200, json=[], complete_qs=True,
            additional_matcher=_requests_matcher_non_existent)
        # Register a requests response for an invalid query
        mocked_requests.register_uri(
            'POST', query.url + query.endpoints['F'], status_code=400, complete_qs=True,
            additional_matcher=_requests_matcher_invalid)

    return query, mocked_requests



"""Vulture whitelist to avoid false positives."""


class Whitelist:
    """Helper class that allows mocking Python objects."""

    def __getattr__(self, _):
        """Mocking magic method __getattr__."""
        pass


whitelist_logging = Whitelist()
whitelist_logging.raiseExceptions

whitelist_cli = Whitelist()
whitelist_cli.run.h

whitelist_transports_clustershell = Whitelist()
whitelist_transports_clustershell.BaseEventHandler.kwargs

whitelist_backends_openstack = Whitelist()
whitelist_backends_openstack.TestOpenStackQuery.test_execute_all.nova_client.return_value.servers.list.side_effect

whitelist_tests_integration_conftest = Whitelist()
whitelist_tests_integration_conftest.pytest_cmdline_preparse
whitelist_tests_integration_conftest.pytest_runtest_makereport

whitelist_tests_integration_test_cli_TestCLI = Whitelist()  # noqa: N816
whitelist_tests_integration_test_cli_TestCLI.setup_method



"""Test external backend module with a GRAMMAR_PREFIX that conflicts with an existing one."""
from cumin.tests.unit.backends.external import ExternalBackendQuery


GRAMMAR_PREFIX = 'D'
""":py:class:`str`: the prefix associate to this grammar, to register this backend into the general grammar.
Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""
query_class = ExternalBackendQuery  # pylint: disable=invalid-name
"""Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""



"""Query handling: factory and builder."""
from pyparsing import ParseException, ParseResults

from cumin import grammar
from cumin.backends import BaseQuery, BaseQueryAggregator, InvalidQueryError


class Query(BaseQueryAggregator):
    """Cumin main query class.

    It has multi-query capability and allow to use a default backend, if set, without additional syntax.
    If a ``default_backend`` is set in the configuration, it will try to execute the query string first with the
    default backend and only if the query is not parsable with that backend it will try to execute it with the
    multi-query grammar.

    When a query is executed, a :py:class:`ClusterShell.NodeSet.NodeSet` with the FQDN of the matched hosts is
    returned.

    Examples:
        >>> import cumin
        >>> from cumin.query import Query
        >>> config = cumin.Config()
        >>> hosts = Query(config).execute(query_string)

    """

    def __init__(self, config):
        """Query constructor, initialize the registered backends.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQueryAggregator.__init__`.

        """
        super().__init__(config)
        external = self.config.get('plugins', {}).get('backends', [])
        self.registered_backends = grammar.get_registered_backends(external=external)
        self.grammar = grammar.grammar(self.registered_backends.keys())

    def execute(self, query_string):
        """Override parent class execute method to implement the multi-query capability.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQueryAggregator.execute`.

        Returns:
            ClusterShell.NodeSet.NodeSet: with the FQDNs of the matching hosts.

        Raises:
            cumin.backends.InvalidQueryError: if unable to parse the query.

        """
        if 'default_backend' not in self.config:
            try:  # No default backend set, using directly the global grammar
                return super().execute(query_string)
            except ParseException as e:
                raise InvalidQueryError(("Unable to parse the query '{query}' with the global grammar and no "
                                         "default backend is set:\n{error}").format(query=query_string, error=e))

        try:  # Default backend set, trying it first
            hosts = self._query_default_backend(query_string)
        except ParseException as e_default:
            try:  # Trying global grammar as a fallback
                hosts = super().execute(query_string)
            except ParseException as e_global:
                raise InvalidQueryError(
                    ("Unable to parse the query '{query}' neither with the default backend '{name}' nor with the "
                     "global grammar:\n{name}: {e_def}\nglobal: {e_glob}").format(
                        query=query_string, name=self.config['default_backend'], e_def=e_default, e_glob=e_global))

        return hosts

    def _query_default_backend(self, query_string):
        """Execute the query with the default backend, according to the configuration.

        Arguments:
            query_string (str): the query string to be parsed and executed with the default backend.

        Returns:
            ClusterShell.NodeSet.NodeSet: with the FQDNs of the matching hosts.

        Raises:
            cumin.backends.InvalidQueryError: if unable to get the default backend from the registered backends.

        """
        for registered_backend in self.registered_backends.values():
            if registered_backend.name == self.config['default_backend']:
                backend = registered_backend
                break
        else:
            raise InvalidQueryError("Default backend '{name}' is not registered: {backends}".format(
                name=self.config['default_backend'], backends=self.registered_backends))

        query = backend.cls(self.config)

        return query.execute(query_string)

    def _parse_token(self, token):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQueryAggregator._parse_token`.

        Raises:
            cumin.backends.InvalidQueryError: on internal parsing error.

        """
        if not isinstance(token, ParseResults):  # pragma: no cover - this should never happen
            raise InvalidQueryError('Expecting ParseResults object, got {type}: {token}'.format(
                type=type(token), token=token))

        token_dict = token.asDict()
        self.logger.trace('Token is: %s', token_dict)

        if self._replace_alias(token_dict):
            return  # This token was an alias and got replaced

        if 'backend' in token_dict and 'query' in token_dict:
            element = self._get_stack_element()
            query = self.registered_backends[token_dict['backend']].cls(self.config)
            element['hosts'] = query.execute(token_dict['query'])
            if 'bool' in token_dict:
                element['bool'] = token_dict['bool']
            self.stack_pointer['children'].append(element)
        elif 'open_subgroup' in token_dict and 'close_subgroup' in token_dict:
            self._open_subgroup()
            if 'bool' in token_dict:
                self.stack_pointer['bool'] = token_dict['bool']
            for subtoken in token:
                if isinstance(subtoken, str):
                    continue
                self._parse_token(subtoken)
            self._close_subgroup()
        else:  # pragma: no cover - this should never happen
            raise InvalidQueryError('Got unexpected token: {token}'.format(token=token))

    def _replace_alias(self, token_dict):
        """Replace any alias in the query in a recursive way, alias can reference other aliases.

        Arguments:
            token_dict (dict): the dictionary of the parsed token returned by the grammar parsing.

        Returns:
            bool: :py:data:`True` if a replacement was made, :py:data`False` otherwise.

        Raises:
            cumin.backends.InvalidQueryError: if unable to replace an alias.

        """
        if 'alias' not in token_dict:
            return False

        alias_name = token_dict['alias']
        if alias_name not in self.config.get('aliases', {}):
            raise InvalidQueryError("Unable to find alias replacement for '{alias}' in the configuration".format(
                alias=alias_name))

        self._open_subgroup()
        if 'bool' in token_dict:
            self.stack_pointer['bool'] = token_dict['bool']

        # Calling BaseQuery._build() directly and not the parent's one to avoid resetting the stack
        BaseQuery._build(self, self.config['aliases'][alias_name])  # pylint: disable=protected-access
        self._close_subgroup()

        return True



"""Colors module."""
from abc import ABCMeta


class ColoredType(ABCMeta):
    """Metaclass to define a new type that dynamically adds static methods to its classes."""

    COLORS = {
        'red': 31,
        'green': 32,
        'yellow': 33,
        'blue': 34,
        'cyan': 36,
    }
    """:py:class:`dict`: a mapping of colors to the ANSI foreground color code."""

    def __getattr__(cls, name):  # noqa: N805 (Prospector reqires an older version of pep8-naming)
        """Dynamically check access to members of classes of this type.

        :Parameters:
            according to Python's Data model :py:meth:`object.__getattr__`.

        """
        color_code = ColoredType.COLORS.get(name, None)
        if color_code is None:
            raise AttributeError("'{cls}' object has no attribute '{attr}'".format(cls=cls.__name__, attr=name))

        return lambda obj: cls._color(color_code, obj)


class Colored(metaclass=ColoredType):
    """Class to manage colored output.

    Available methods are dynamically added based on the keys of the :py:const:`ColoredType.COLORS` dictionary.
    For each color a method with the color name is available to color any object with that specific color code.

    Examples::

        Colored.green(object)

    """

    disabled = False
    """:py:class:`bool`: switch to globally control the coloring. Set it to :py:const`True` to disable all coloring."""

    @staticmethod
    def _color(color_code, obj):
        """Color the given object, unless coloring is globally disabled.

        Arguments:
            color_code (int): a valid ANSI escape sequence color code.
            obj (mixed): the object to color.

        Return:
            str: the string representation of the object encapsulated in the red ANSI escape sequence.

        """
        message = str(obj)

        if not message:
            return ''

        if Colored.disabled:
            return message

        return '\x1b[{code}m{message}\x1b[39m'.format(code=color_code, message=message)



"""Transport tests."""
# pylint: disable=protected-access,no-self-use
from unittest import mock

import pytest

from ClusterShell.NodeSet import NodeSet

import cumin  # noqa: F401 (dynamically used in TestCommand)

from cumin import transports
from cumin.transports import ProgressBars


class ConcreteBaseWorker(transports.BaseWorker):
    """Concrete class for BaseWorker."""

    def execute(self):
        """Required by BaseWorker."""

    def get_results(self):
        """Required by BaseWorker."""
        yield "node", "output"

    @property
    def handler(self):
        """Required by BaseWorker."""
        return self._handler

    @handler.setter
    def handler(self, value):
        """Required by BaseWorker."""
        self._handler = value


class Commands:
    """Helper class to define a list of commands to test."""

    command_with_options = r'command --with "options" -a -n -d params with\ spaces'
    command_with_options_equivalent = r"command  --with  'options'  -a  -n  -d  params  with\ spaces"
    command_with_nested_quotes = 'command --with \'nested "quotes"\' -a -n -d params with\\ spaces'
    different_command = r"command  --with  'other options'  -a  -n  -d  other_params  with\ spaces"

    def __init__(self):
        """Initialize test commands."""
        self.commands = [
            {'command': 'command1'},
            {'command': 'command1', 'timeout': 5},
            {'command': 'command1', 'ok_codes': [0, 255]},
            {'command': 'command1', 'timeout': 5, 'ok_codes': [0, 255]},
            {'command': self.command_with_options},
            {'command': self.command_with_options, 'timeout': 5},
            {'command': self.command_with_options, 'ok_codes': [0, 255]},
            {'command': self.command_with_options, 'timeout': 5, 'ok_codes': [0, 255]},
            {'command': self.command_with_nested_quotes},
            {'command': self.command_with_nested_quotes, 'timeout': 5},
            {'command': self.command_with_nested_quotes, 'ok_codes': [0, 255]},
            {'command': self.command_with_nested_quotes, 'timeout': 5, 'ok_codes': [0, 255]},
        ]

        for command in self.commands:
            command['obj'] = transports.Command(
                command['command'], timeout=command.get('timeout', None), ok_codes=command.get('ok_codes', None))


@pytest.mark.parametrize('command', Commands().commands)
class TestCommandParametrized:
    """Command class tests executed for each parametrized command."""

    def test_instantiation(self, command):
        """A new Command instance should set the command property to the given command."""
        assert isinstance(command['obj'], transports.Command)
        assert command['obj'].command == command['command']
        assert command['obj']._timeout == command.get('timeout', None)
        assert command['obj']._ok_codes == command.get('ok_codes', None)

    def test_repr(self, command):
        """A repr of a Command should allow to instantiate an instance with the same properties."""
        # Bandit and pylint would require to use ast.literal_eval, but it will not work with objects
        command_repr = repr(command['obj'])
        if r'\ ' in command_repr:
            return  # Skip tests with bash-escaped spaces are they will trigger DeprecationWarning

        command_instance = eval(command_repr)  # nosec pylint: disable=eval-used
        assert isinstance(command_instance, transports.Command)
        assert repr(command_instance) == repr(command['obj'])
        assert command_instance.command == command['obj'].command
        assert command_instance._timeout == command['obj']._timeout
        assert command_instance._ok_codes == command['obj']._ok_codes

    def test_str(self, command):
        """A cast to string of a Command should return its command."""
        assert str(command['obj']) == command['command']

    def test_eq(self, command):
        """A Command instance can be compared to another or to a string with the equality operator."""
        assert command['obj'] == transports.Command(
            command['command'], timeout=command.get('timeout', None), ok_codes=command.get('ok_codes', None))

        if command.get('timeout', None) is None and command.get('ok_codes', None) is None:
            assert command['obj'] == command['command']

        with pytest.raises(ValueError, match='Unable to compare instance of'):
            command['obj'] == 1  # pylint: disable=pointless-statement

    def test_ne(self, command):
        """A Command instance can be compared to another or to a string with the inequality operator."""
        # Different command with same or differnt properties
        assert command['obj'] != transports.Command(
            Commands.different_command, timeout=command.get('timeout', None),
            ok_codes=command.get('ok_codes', None))
        assert command['obj'] != transports.Command(
            Commands.different_command, timeout=999, ok_codes=command.get('ok_codes', None))
        assert command['obj'] != transports.Command(
            Commands.different_command, timeout=command.get('timeout', None), ok_codes=[99])
        assert command['obj'] != transports.Command(Commands.different_command, timeout=999, ok_codes=[99])
        assert command['obj'] != Commands.different_command

        # Same command, properties different
        assert command['obj'] != transports.Command(
            command['command'], timeout=999, ok_codes=command.get('ok_codes', None))
        assert command['obj'] != transports.Command(
            command['command'], timeout=command.get('timeout', None), ok_codes=[99])
        assert command['obj'] != transports.Command(command['command'], timeout=999, ok_codes=[99])

        if command.get('timeout', None) is not None or command.get('ok_codes', None) is not None:
            assert command['obj'] != command['command']

        with pytest.raises(ValueError, match='Unable to compare instance of'):
            command['obj'] == 1  # pylint: disable=pointless-statement

    def test_timeout_getter(self, command):
        """Should return the timeout set, None otherwise."""
        if command['obj'].timeout is not None and command['obj']._timeout is not None:
            assert command['obj'].timeout == pytest.approx(command['obj']._timeout)

    def test_ok_codes_getter(self, command):
        """Should return the ok_codes set, [0] otherwise."""
        assert command['obj'].ok_codes == command.get('ok_codes', [0])


class TestCommand:
    """Command class non parametrized tests."""

    def test_eq_equivalent(self):
        """Two Commadn instances with equivalent comamnds just formatted differently should be considered equal."""
        command1 = transports.Command(Commands.command_with_options)
        command2 = transports.Command(Commands.command_with_options_equivalent)
        assert command1 == command2

    def test_timeout_setter(self):
        """Should set the timeout to its value, converted to float if integer. Unset it if None is passed."""
        command = transports.Command('command1')
        command.timeout = 1.0
        assert command._timeout == pytest.approx(1.0)
        command.timeout = None
        assert command._timeout is None
        command.timeout = 1
        assert command._timeout == pytest.approx(1.0)
        with pytest.raises(transports.WorkerError, match='timeout must be a positive float'):
            command.timeout = -1.0

    def test_ok_codes_getter_empty(self):
        """Should return the ok_codes set, [0] otherwise."""
        # Test empty list
        command = transports.Command('command1')
        assert command.ok_codes == [0]
        command.ok_codes = []
        assert command.ok_codes == []
        command.ok_codes = [1, 255]
        assert command.ok_codes == [1, 255]

    def test_ok_codes_setter(self):
        """Should set the ok_codes to its value, unset it if None is passed."""
        command = transports.Command('command1')
        assert command._ok_codes is None
        for i in range(256):
            codes = [i]
            command.ok_codes = codes
            assert command._ok_codes == codes
            codes.insert(0, 0)
            command.ok_codes = codes
            assert command._ok_codes == codes

        command.ok_codes = None
        assert command._ok_codes is None

        command.ok_codes = []
        assert command._ok_codes == []

        with pytest.raises(transports.WorkerError, match='ok_codes must be a list'):
            command.ok_codes = 'invalid_value'

        message = 'must be a list of integers in the range'
        for i in (-1, 0.0, 100.0, 256, 'invalid_value'):
            codes = [i]
            with pytest.raises(transports.WorkerError, match=message):
                command.ok_codes = codes

            codes.insert(0, 0)
            with pytest.raises(transports.WorkerError, match=message):
                command.ok_codes = codes


class TestState:
    """State class tests."""

    def test_instantiation_no_init(self):
        """A new State without an init value should start in the pending state."""
        state = transports.State()
        assert state._state == transports.State.pending

    def test_instantiation_init_ok(self):
        """A new State with a valid init value should start in this state."""
        state = transports.State(init=transports.State.running)
        assert state._state == transports.State.running

    def test_instantiation_init_ko(self):
        """A new State with an invalid init value should raise InvalidStateError."""
        with pytest.raises(transports.InvalidStateError, match='is not a valid state'):
            transports.State(init='invalid_state')

    def test_getattr_current(self):
        """Accessing the 'current' property should return the current state."""
        assert transports.State().current == transports.State.pending

    def test_getattr_is_valid_state(self):
        """Accessing a property named is_{a_valid_state_name} should return a boolean."""
        state = transports.State(init=transports.State.failed)
        assert not state.is_pending
        assert not state.is_scheduled
        assert not state.is_running
        assert not state.is_timeout
        assert not state.is_success
        assert state.is_failed

    def test_getattr_invalid_property(self):
        """Accessing a property with an invalid name should raise AttributeError."""
        state = transports.State(init=transports.State.failed)
        with pytest.raises(AttributeError, match='object has no attribute'):
            state.invalid_property  # pylint: disable=pointless-statement

    def test_repr(self):
        """A State repr should return its representation that allows to recreate the same State instance."""
        assert repr(transports.State()) == 'cumin.transports.State(init={state})'.format(state=transports.State.pending)
        state = transports.State.running
        assert repr(transports.State(init=state)) == 'cumin.transports.State(init={state})'.format(state=state)

    def test_str(self):
        """A State string should return its string representation."""
        assert str(transports.State()) == 'pending'
        assert str(transports.State(init=transports.State.running)) == 'running'

    def test_cmp_state(self):
        """Two State instance can be compared between each other."""
        state = transports.State()
        greater_state = transports.State(init=transports.State.failed)
        same_state = transports.State()

        assert greater_state > state
        assert greater_state >= state
        assert same_state >= state
        assert state < greater_state
        assert state <= greater_state
        assert state <= same_state
        assert state == same_state
        assert state != greater_state

    def test_cmp_int(self):
        """A State instance can be compared with integers."""
        state = transports.State()
        greater_state = transports.State.running
        same_state = transports.State.pending

        assert greater_state > state
        assert greater_state >= state
        assert same_state >= state
        assert state < greater_state
        assert state <= greater_state
        assert state <= same_state
        assert state == same_state
        assert state != greater_state

    def test_cmp_invalid(self):
        """Trying to compare a State instance with an invalid object should raise ValueError."""
        state = transports.State()
        invalid_state = 'invalid_state'
        with pytest.raises(ValueError, match='Unable to compare instance'):
            state == invalid_state  # pylint: disable=pointless-statement

    def test_update_invalid_state(self):
        """Trying to update a State with an invalid value should raise ValueError."""
        state = transports.State()
        with pytest.raises(ValueError, match='State must be one of'):
            state.update('invalid_state')

    def test_update_invalid_transition(self):
        """Trying to update a State with an invalid transition should raise StateTransitionError."""
        state = transports.State()
        with pytest.raises(transports.StateTransitionError, match='the allowed states are'):
            state.update(transports.State.failed)

    def test_update_ok(self):
        """Properly updating a State should update it without errors."""
        state = transports.State()
        state.update(transports.State.scheduled)
        assert state.current == transports.State.scheduled
        state.update(transports.State.running)
        assert state.current == transports.State.running
        state.update(transports.State.success)
        assert state.current == transports.State.success
        state.update(transports.State.pending)
        assert state.current == transports.State.pending


class TestTarget:
    """Target class tests."""

    def setup_method(self, _):
        """Initialize default properties and instances."""
        # pylint: disable=attribute-defined-outside-init
        self.hosts_list = ['host' + str(i) for i in range(10)]
        self.hosts = cumin.nodeset_fromlist(self.hosts_list)

    def test_init_no_hosts(self):
        """Creating a Target instance with empty hosts should raise WorkerError."""
        with pytest.raises(transports.WorkerError, match="must be a non-empty ClusterShell NodeSet or list"):
            transports.Target([])

    def test_init_nodeset(self):
        """Creating a Target instance with a NodeSet and without optional parameter should return their defaults."""
        target = transports.Target(self.hosts)
        assert target.hosts == self.hosts
        assert target.batch_size == len(self.hosts)
        assert target.batch_sleep == 0.0

    def test_init_list(self):
        """Creating a Target instance with a list and without optional parameter should return their defaults."""
        target = transports.Target(self.hosts_list)
        assert target.hosts == self.hosts
        assert target.batch_size == len(self.hosts)
        assert target.batch_sleep == 0.0

    def test_init_invalid(self):
        """Creating a Target instance with invalid hosts should raise WorkerError."""
        with pytest.raises(transports.WorkerError, match="must be a non-empty ClusterShell NodeSet or list"):
            transports.Target(set(self.hosts_list))

    @mock.patch('cumin.transports.logging.Logger.debug')
    def test_init_batch_size(self, mocked_logger):
        """Creating a Target instance with a batch_size should set it to it's value, if valid."""
        target = transports.Target(self.hosts, batch_size=5)
        assert target.batch_size == 5

        target = transports.Target(self.hosts, batch_size=len(self.hosts) + 1)
        assert target.batch_size == len(self.hosts)
        assert mocked_logger.called

        target = transports.Target(self.hosts, batch_size=None)
        assert target.batch_size == len(self.hosts)

        with pytest.raises(transports.WorkerError, match='must be a positive integer'):
            transports.Target(self.hosts, batch_size=0)

    def test_init_batch_size_perc(self):
        """Creating a Target instance with a batch_size_ratio should set batch_size to the appropriate value."""
        target = transports.Target(self.hosts, batch_size_ratio=0.5)
        assert target.batch_size == 5

        target = transports.Target(self.hosts, batch_size_ratio=1.0)
        assert target.batch_size == len(self.hosts)

        target = transports.Target(self.hosts, batch_size_ratio=None)
        assert target.batch_size == len(self.hosts)

        with pytest.raises(transports.WorkerError, match='parameters are mutually exclusive'):
            transports.Target(self.hosts, batch_size=1, batch_size_ratio=0.5)

        with pytest.raises(transports.WorkerError, match='has generated a batch_size of 0 hosts'):
            transports.Target(self.hosts, batch_size_ratio=0.0)

    @pytest.mark.parametrize('ratio', (1, 2.0, -0.1))
    def test_init_batch_size_perc_range(self, ratio):
        """Creating a Target instance with an invalid batch_size_ratio should raise WorkerError."""
        with pytest.raises(transports.WorkerError, match='must be a float between 0.0 and 1.0'):
            transports.Target(self.hosts, batch_size_ratio=ratio)

    def test_init_batch_sleep(self):
        """Creating a Target instance with a batch_sleep should set it to it's value, if valid."""
        target = transports.Target(self.hosts, batch_sleep=5.0)
        assert target.batch_sleep == pytest.approx(5.0)

        target = transports.Target(self.hosts, batch_sleep=None)
        assert target.batch_sleep == pytest.approx(0.0)

        with pytest.raises(transports.WorkerError):
            transports.Target(self.hosts, batch_sleep=0)

        with pytest.raises(transports.WorkerError):
            transports.Target(self.hosts, batch_sleep=-1.0)

    def test_first_batch(self):
        """The first_batch property should return the first_batch of hosts."""
        size = 5
        target = transports.Target(self.hosts, batch_size=size)
        assert len(target.first_batch) == size
        assert target.first_batch == cumin.nodeset_fromlist(self.hosts[:size])
        assert isinstance(target.first_batch, NodeSet)


class TestBaseWorker:
    """Concrete BaseWorker class for tests."""

    def test_instantiation(self):
        """Raise if instantiated directly, should return an instance of BaseWorker if inherited."""
        target = transports.Target(cumin.nodeset('node1'))
        with pytest.raises(TypeError):
            transports.BaseWorker({}, target)  # pylint: disable=abstract-class-instantiated

        assert isinstance(ConcreteBaseWorker({}, transports.Target(cumin.nodeset('node[1-2]'))), transports.BaseWorker)

    @mock.patch.dict(transports.os.environ, {}, clear=True)
    def test_init(self):
        """Constructor should save config and set environment variables."""
        env_dict = {'ENV_VARIABLE': 'env_value'}
        config = {'transport': 'test_transport',
                  'environment': env_dict}

        assert transports.os.environ == {}
        worker = ConcreteBaseWorker(config, transports.Target(cumin.nodeset('node[1-2]')))
        assert transports.os.environ == env_dict
        assert worker.config == config


class TestConcreteBaseWorker:
    """BaseWorker test class."""

    def setup_method(self, _):
        """Initialize default properties and instances."""
        # pylint: disable=attribute-defined-outside-init
        self.worker = ConcreteBaseWorker({}, transports.Target(cumin.nodeset('node[1-2]')))
        self.commands = [transports.Command('command1'), transports.Command('command2')]

    def test_commands_getter(self):
        """Access to commands property should return an empty list if not set and the list of commands otherwise."""
        assert self.worker.commands == []
        self.worker._commands = self.commands
        assert self.worker.commands == self.commands
        self.worker._commands = None
        assert self.worker.commands == []

    def test_commands_setter(self):
        """Raise WorkerError if trying to set it not to a list, set it otherwise."""
        with pytest.raises(transports.WorkerError, match='commands must be a list'):
            self.worker.commands = 'invalid_value'

        with pytest.raises(transports.WorkerError, match='commands must be a list of Command objects'):
            self.worker.commands = [1, 'command2']

        self.worker.commands = self.commands
        assert self.worker._commands == self.commands
        self.worker.commands = None
        assert self.worker._commands is None
        self.worker.commands = []
        assert self.worker._commands == []
        self.worker.commands = ['command1', 'command2']
        assert self.worker._commands == self.commands

    def test_timeout_getter(self):
        """Return default value if not set, the value otherwise."""
        assert self.worker.timeout == 0
        self.worker._timeout = 10
        assert self.worker.timeout == 10

    def test_timeout_setter(self):
        """Raise WorkerError if not a positive integer, set it otherwise."""
        message = r'timeout must be a positive integer'
        with pytest.raises(transports.WorkerError, match=message):
            self.worker.timeout = -1

        with pytest.raises(transports.WorkerError, match=message):
            self.worker.timeout = 0

        self.worker.timeout = 10
        assert self.worker._timeout == 10
        self.worker.timeout = None
        assert self.worker._timeout is None

    def test_success_threshold_getter(self):
        """Return default value if not set, the value otherwise."""
        assert self.worker.success_threshold == pytest.approx(1.0)
        for success_threshold in (0.0, 0.0001, 0.5, 0.99):
            self.worker._success_threshold = success_threshold
            assert self.worker.success_threshold == pytest.approx(success_threshold)

    def test_success_threshold_setter(self):
        """Raise WorkerError if not float between 0 and 1, set it otherwise."""
        message = r'success_threshold must be a float beween 0 and 1'
        with pytest.raises(transports.WorkerError, match=message):
            self.worker.success_threshold = 1

        with pytest.raises(transports.WorkerError, match=message):
            self.worker.success_threshold = -0.1

        self.worker.success_threshold = 0.3
        assert self.worker._success_threshold == pytest.approx(0.3)


class TestModuleFunctions:
    """Transports module functions test class."""

    def test_validate_list(self):
        """Should raise a WorkerError if the argument is not a list."""
        transports.validate_list('Test', ['value1'])
        transports.validate_list('Test', ['value1', 'value2'])
        transports.validate_list('Test', [], allow_empty=True)

        with pytest.raises(transports.WorkerError, match=r'Test must be a non-empty list'):
            transports.validate_list('Test', [])

        message = r'Test must be a list'
        for invalid_value in (0, None, 'invalid_value', {'invalid': 'value'}):
            with pytest.raises(transports.WorkerError, match=message):
                transports.validate_list('Test', invalid_value)

    def test_validate_positive_integer(self):
        """Should raise a WorkerError if the argument is not a positive integer or None."""
        transports.validate_positive_integer('Test', None)
        transports.validate_positive_integer('Test', 1)
        transports.validate_positive_integer('Test', 100)

        message = r'Test must be a positive integer'
        for invalid_value in (0, -1, 'invalid_value', ['invalid_value']):
            with pytest.raises(transports.WorkerError, match=message):
                transports.validate_positive_integer('Test', invalid_value)

    def test_raise_error(self):
        """Should raise a WorkerError."""
        with pytest.raises(transports.WorkerError, match='Test message'):
            transports.raise_error('Test', 'message', 'value')


@mock.patch('cumin.transports.tqdm')
class TestProgressBars:
    """A class that tests ProgressBars."""

    def test_init_intialize_progress_bars_with_correct_size(self, tqdm):
        """Progress bars are initialized at the correct size."""
        progress = ProgressBars()
        progress.init(10)

        assert tqdm.call_count == 2
        _, kwargs = tqdm.call_args
        assert kwargs['total'] == 10

    def test_progress_bars_are_closed(self, tqdm):  # pylint: disable=unused-argument
        """Progress bars are closed."""
        progress = ProgressBars()
        progress.init(10)

        progress.close()

        assert progress.pbar_ok.close.called  # pylint: disable=no-member
        assert progress.pbar_ko.close.called  # pylint: disable=no-member

    def test_progress_bar_is_updated_on_success(self, tqdm):  # pylint: disable=unused-argument
        """Progress bar is updated on success."""
        progress = ProgressBars()
        progress.init(10)

        progress.update_success(5)

        assert progress.pbar_ok.update.called_once_with(5)  # pylint: disable=no-member

    def test_progress_bar_is_updated_on_failure(self, tqdm):  # pylint: disable=unused-argument
        """Progress bar is updated on failure."""
        progress = ProgressBars()
        progress.init(10)

        progress.update_failed(3)

        assert progress.pbar_ko.update.called_once_with(3)  # pylint: disable=no-member



"""PuppetDB backend."""

from re import IGNORECASE
from string import capwords

import pyparsing as pp
import requests

from requests.packages import urllib3

from cumin import nodeset, nodeset_fromlist
from cumin.backends import BaseQuery, InvalidQueryError


CATEGORIES = ('C', 'F', 'O', 'P', 'R')
""":py:func:`tuple`: available categories in the grammar.

* ``C``: shortcut for querying resources of type ``Class``, equivalent of `R:Class = class_path``.
* ``F``: for querying facts.
* ``O``: shortcut for querying resources of type ``Class`` that starts with ``Role::``.
* ``P``: shortcut for querying resources of type ``Class`` that starts with ``Profile::``.
* ``R``: for querying generic resources.
"""

OPERATORS = ('=', '>=', '<=', '<', '>', '~')
""":py:func:`tuple`: available operators in the grammar, the same available in PuppetDB API.

The ``~`` one is used for regex matching.
"""


class ParsedString:
    """Simple string wrapper which can communicate if a string should be enquoted downstream."""

    def __init__(self, string, is_quoted):
        """Constructor for ParsedString.

        Arguments:
           string (str): The string to store in this object.
           is_quoted (bool): Whether the output should be quoted when this is converted to a string.

        """
        self.string = str(string)
        self.is_quoted = is_quoted

    def __str__(self):
        """Return a string version of this value, enquoted or not based on the is_quoted property."""
        if self.is_quoted:
            return '"{}"'.format(self.string)
        return self.string

    def capwords(self, sep):
        """Perform capwords operation on internal value and return a new ParsedString.

        :Parameters:
            according to :py:meth:`string.capwords`.

        """
        return ParsedString(capwords(self.string, sep), self.is_quoted)

    def replace(self, old, new, count=-1):
        """Perform replace operation on internal value and return a new ParsedString.

        :Parameters:
            according to :py:meth:`str.replace`.
        """
        return ParsedString(self.string.replace(old, new, count), self.is_quoted)


def grammar():  # pylint: disable=too-many-locals
    """Define the query grammar.

    Backus-Naur form (BNF) of the grammar::

            <grammar> ::= <item> | <item> <and_or> <grammar>
               <item> ::= [<neg>] <query-token> | [<neg>] "(" <grammar> ")"
        <query-token> ::= <token> | <hosts>
              <token> ::= <category>:<key> [<operator> <value>]
              <value> ::= <numeric> | <bareword> | <quoted_string> | <unquoted_string>

    Given that the pyparsing library defines the grammar in a BNF-like style, for the details of the tokens not
    specified above check directly the source code.

    Returns:
        pyparsing.ParserElement: the grammar parser.

    """
    # Boolean operators
    and_or = (pp.CaselessKeyword('and') | pp.CaselessKeyword('or'))('bool')
    # 'neg' is used as label to allow the use of dot notation, 'not' is a reserved word in Python
    neg = pp.CaselessKeyword('not')('neg')

    operator = pp.oneOf(OPERATORS, caseless=True)('operator')  # Comparison operators
    quoted_string = pp.quotedString.copy().addParseAction(pp.removeQuotes)  # Both single and double quotes are allowed

    # Parentheses
    lpar = pp.Literal('(')('open_subgroup')
    rpar = pp.Literal(')')('close_subgroup')

    # Hosts selection: glob (*) and clustershell (,!&^[]) syntaxes are allowed:
    # i.e. host10[10-42].*.domain
    hosts = quoted_string | (~(and_or | neg) + pp.Word(pp.alphanums + '-_.*,!&^[]'))

    # Key-value token for allowed categories using the available comparison operators
    # i.e. F:key = value
    category = pp.oneOf(CATEGORIES, caseless=True)('category')
    key = pp.Word(pp.alphanums + '-_.%@:')('key')
    selector = pp.Combine(category + ':' + key)  # i.e. F:key
    # All printables characters except the parentheses that are part of this or the global grammar
    all_but_par = ''.join([c for c in pp.printables if c not in ('(', ')', '{', '}')])

    # PuppetDB accepts JSON Atoms
    bareword = pp.oneOf(('true', 'false'))

    # octal numbers are bare numerics that lead with 0.
    octal = pp.Word("0", "01234567", min=2).addParseAction(lambda toks: int(toks[0], 8))
    # hex integers are in the format 0x[0-9A-F]+
    hexadecimal = pp.Regex(r'0x[0-9A-F]+', flags=IGNORECASE).addParseAction(lambda toks: int(toks[0], 16))
    number = pp.pyparsing_common.number

    # label indicates post-processing needed (value = nonquoted, quoted=quoted)
    value = (hexadecimal ^ octal ^ number ^ bareword)('value') ^ (quoted_string ^ pp.Word(all_but_par))('quoted')

    token = selector + pp.Optional(operator + value)

    # Final grammar, see the docstring for its BNF based on the tokens defined above
    # Groups are used to split the parsed results for an easy access
    full_grammar = pp.Forward()
    item = pp.Group(pp.Optional(neg) + (token | hosts('hosts'))) | pp.Group(
        pp.Optional(neg) + lpar + full_grammar + rpar)
    full_grammar << item + pp.ZeroOrMore(pp.Group(and_or) + full_grammar)  # pylint: disable=expression-not-assigned

    return full_grammar


class PuppetDBQuery(BaseQuery):
    """PuppetDB query builder.

    The `puppetdb` backend allow to use an existing PuppetDB instance for the hosts selection.
    The supported PuppetDB API versions are 3 and 4. It can be specified via the api_version configuration key, if
    not configured, the v4 will be used.

    * Each query part can be composed with the others using boolean operators (``and``, ``or``, ``not``)
    * Multiple query parts can be grouped together with parentheses (``(``, ``)``).
    * A query part can be of two different types:

      * ``Hostname matching``: this is a simple string that be used to match directly the hostname of the hosts in the
        selected backend. It allows for glob expansion (``*``) and the use of the powerful
        :py:class:`ClusterShell.NodeSet.NodeSet`.
      * ``Category matching``: an identifier composed by a category, a colon and a key, followed by a comparison
        operator and a value, as in ``F:key = value``.

    * Values may be of various types supported by PuppetDB (numerics, boolean, and strings) for example:

      * Booleans: ``true``, ``false``
      * Strings: ``'a string'``, and unquoted single words that aren't ``true`` or ``false`` and do not start with an
        integer.
      * Numeric values: ``15``, ``23.5``, ``0``, ``0xfa`` *Note: hexadecimal and octal numbers are supported by cumin
        but converted into normal integers. Some fields in PuppetDB may have hex or octal stored as strings, and should
        be quoted such as* ``'0xfa'``.

      *Note: PuppetDB may or may not support a particular value type for a particular resource.*

    Some query examples:

    * All hosts: ``*``
    * Hosts globbing: ``host10*``
    * :py:class:`ClusterShell.NodeSet.NodeSet` syntax for hosts expansion: ``host10[10-42].domain``
    * Category based key-value selection:

      * ``R:Resource::Name``: query all the hosts that have a resource of type `Resource::Name`.
      * ``R:Resource::Name = 'resource-title'``: query all the hosts that have a resource of type `Resource::Name`
        whose title is ``resource-title``. For example ``R:Class = MyModule::MyClass``.
      * ``R:Resource::Name@field = 'some-value'``: query all the hosts that have a resource of type ``Resource::Name``
        whose field ``field`` has the value ``some-value``. The valid fields are: ``tag``, ``certname``, ``type``,
        ``title``, ``exported``, ``file``, ``line``. The previous syntax is a shortcut for this one with the field
        ``title``.
      * ``R:Resource::Name%param = 'some-value'``: query all the hosts that have a resource of type ``Resource::Name``
        whose parameter ``param`` has the value ``some-value``.
      * ``C:Class::Name``: special shortcut to query all the hosts that have a resource of type ``Class`` whose name
        is ``Class::Name``. The ``Class::Name`` part is completely arbitrary and depends on the puppet hierarchy
        chosen. It's equivalent to ``R:Class = Class::Name``, with the addition that the ``param`` and ``field``
        selectors described above can be used directly without the need to add another condition.
      * ``O:Module::Name``: special shortcut to query all the hosts that have a resource of type ``Class`` whose name
        is ``Role::Module::Name``. The ``Module::Name`` part is completely arbitrary and depends on the puppet
        hierarchy chosen. It's equivalent to ``R:Class = Role::Module::Name``, with the addition that the ``param`` and
        ``field`` selectors described above can be used directly without the need to add another condition, although
        usually roles should not have parameters in the role/profile Puppet paradigm.
      * ``P:Module::Name``: special shortcut to query all the hosts that have a resource of type ``Class`` whose name
        is ``Profile::Module::Name``. The ``Module::Name`` part is completely arbitrary and depends on the puppet
        hierarchy chosen. It's equivalent to ``R:Class = Profile::Module::Name``, with the addition that the ``param``
        and ``field`` selectors described above can be used directly without the need to add another condition.
      * ``F:FactName = value``: query all the hosts that have a fact ``FactName``, as reported by facter, with the
        value ``value``.
      * Mixed facts/resources queries are not supported, but the same result can be achieved using the main grammar
        with multiple subqueries for the PuppetDB backend.

    * All hosts with physicalcorecount fact greater than 2: ``F:physicalcorecount > 2``
    * A complex selection for facts:
      ``host10[10-42].*.domain or (not F:key1 = value1 and host10*) or (F:key2 > value2 and F:key3 ~ '^value[0-9]+')``
    """

    base_url_template = '{scheme}://{host}:{port}'
    """:py:class:`str`: string template in the :py:meth:`str.format` style used to generate the base URL of the
    PuppetDB server."""

    endpoints = {'C': 'resources', 'F': 'nodes', 'O': 'resources', 'P': 'resources', 'R': 'resources'}
    """:py:class:`dict`: dictionary with the mapping of the available categories in the grammar to the PuppetDB API
    endpoints."""

    category_prefixes = {'C': '', 'O': 'Role', 'P': 'Profile'}
    """:py:class:`dict`: dictionary with the mapping of special categories to title prefixes."""

    grammar = grammar()
    """:py:class:`pyparsing.ParserElement`: load the grammar parser only once in a singleton-like way."""

    def __init__(self, config):
        """Query constructor for the PuppetDB backend.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery.__init__`.

        """
        super().__init__(config)
        self.grouped_tokens = None
        self.current_group = self.grouped_tokens
        self._endpoint = None
        puppetdb_config = self.config.get('puppetdb', {})
        base_url = self.base_url_template.format(
            scheme=puppetdb_config.get('scheme', 'https'),
            host=puppetdb_config.get('host', 'localhost'),
            port=puppetdb_config.get('port', 443))

        self.api_version = puppetdb_config.get('api_version', 4)
        if self.api_version == 3:
            self.url = base_url + '/v3/'
            self.hosts_keys = {'nodes': 'name', 'resources': 'certname'}
        elif self.api_version == 4:
            self.url = base_url + '/pdb/query/v4/'
            self.hosts_keys = {'nodes': 'certname', 'resources': 'certname'}
        else:
            raise InvalidQueryError('Unsupported PuppetDB API version {ver}'.format(ver=self.api_version))

        for exception in puppetdb_config.get('urllib3_disable_warnings', []):
            urllib3.disable_warnings(category=getattr(urllib3.exceptions, exception))

    @property
    def endpoint(self):
        """Endpoint in the PuppetDB API for the current query.

        :Getter:
            Returns the current `endpoint` or a default value if not set.

        :Setter:
            :py:class:`str`: the value to set the `endpoint` to.

        Raises:
            cumin.backends.InvalidQueryError: if trying to set it to an invalid `endpoint` or mixing endpoints in a
                single query.

        """
        return self._endpoint or 'nodes'

    @endpoint.setter
    def endpoint(self, value):
        """Setter for the `endpoint` property. The relative documentation is in the getter."""
        if value not in self.endpoints.values():
            raise InvalidQueryError("Invalid value '{endpoint}' for endpoint property".format(endpoint=value))
        if self._endpoint is not None and value != self._endpoint:
            raise InvalidQueryError('Mixed endpoints are not supported, use the global grammar to mix them.')

        self._endpoint = value

    def _open_subgroup(self):
        """Handle subgroup opening."""
        token = PuppetDBQuery._get_grouped_tokens()
        token['parent'] = self.current_group
        self.current_group['tokens'].append(token)
        self.current_group = token

    def _close_subgroup(self):
        """Handle subgroup closing."""
        self.current_group = self.current_group['parent']

    @staticmethod
    def _get_grouped_tokens():
        """Return an empty grouped tokens structure.

        Returns:
            dict: the dictionary with the empty grouped tokens structure.

        """
        return {'parent': None, 'bool': None, 'tokens': []}

    def _build(self, query_string):
        """Override parent class _build method to reset tokens and add logging.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._build`.

        """
        self.grouped_tokens = PuppetDBQuery._get_grouped_tokens()
        self.current_group = self.grouped_tokens
        super()._build(query_string)
        self.logger.trace('Query tokens: %s', self.grouped_tokens)

    def _execute(self):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._execute`.

        Returns:
            ClusterShell.NodeSet.NodeSet: with the FQDNs of the matching hosts.

        """
        query = self._get_query_string(group=self.grouped_tokens).format(host_key=self.hosts_keys[self.endpoint])
        hosts = self._api_call(query)
        unique_hosts = nodeset_fromlist([host[self.hosts_keys[self.endpoint]] for host in hosts])
        self.logger.debug("Queried puppetdb for '%s', got '%d' results.", query, len(unique_hosts))

        return unique_hosts

    def _add_category(self, category, key, value=None, operator='=', neg=False):
        """Add a category token to the query 'F:key = value'.

        Arguments:
            category (str): the category of the token, one of :py:const:`CATEGORIES`.
            key (str): the key for this category.
            value (str, optional): the value to match, if not specified the key itself will be matched.
            operator (str, optional): the comparison operator to use, one of :py:const:`OPERATORS`.
            neg (bool, optional): whether the token must be negated.

        Raises:
            cumin.backends.InvalidQueryError: on internal parsing error.

        """
        self.endpoint = self.endpoints[category]
        if operator == '~':
            # PuppetDB API requires to escape every backslash
            # See: https://puppet.com/docs/puppetdb/4.4/api/query/v4/ast.html#regexp-match
            value = value.replace('\\', '\\\\')
        if category in ('C', 'O', 'P'):
            query = self._get_special_resource_query(category, key, value, operator)
        elif category == 'R':
            query = self._get_resource_query(key, value, operator)
        elif category == 'F':
            query = '["{op}", ["fact", "{key}"], {val}]'.format(op=operator, key=key, val=value)
        else:  # pragma: no cover - this should never happen
            raise InvalidQueryError(
                "Got invalid category '{category}', one of F|O|P|R expected".format(category=category))

        if neg:
            query = '["not", {query}]'.format(query=query)

        self.current_group['tokens'].append(query)

    def _add_hosts(self, hosts, neg=False):
        """Add a list of hosts to the query.

        Arguments:
            hosts (ClusterShell.NodeSet.NodeSet): with the list of hosts to search.
            neg (bool, optional): whether the token must be negated.

        """
        if not hosts:
            return

        hosts_tokens = []
        for host in hosts:
            operator = '='
            # Convert a glob expansion into a regex
            if '*' in host:
                operator = '~'
                host = r'^' + host.replace('.', r'\\.').replace('*', '.*') + r'$'

            hosts_tokens.append('["{op}", "{{host_key}}", "{host}"]'.format(op=operator, host=host))

        query = '["or", {hosts}]'.format(hosts=', '.join(hosts_tokens))
        if neg:
            query = '["not", {query}]'.format(query=query)

        self.current_group['tokens'].append(query)

    def _parse_token(self, token):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._parse_token`.

        Raises:
            cumin.backends.InvalidQueryError: on internal parsing error.

        """
        if isinstance(token, str):
            return

        token_dict = token.asDict()
        # post-process types
        if 'quoted' in token_dict:
            token_dict['value'] = ParsedString(token_dict['quoted'], True)
            del token_dict['quoted']
        elif 'value' in token_dict:
            token_dict['value'] = ParsedString(token_dict['value'], False)

        # Based on the token type build the corresponding query object
        if 'open_subgroup' in token_dict:
            self._open_subgroup()
            for subtoken in token:
                self._parse_token(subtoken)
            self._close_subgroup()

        elif 'bool' in token_dict:
            self._add_bool(token_dict['bool'])

        elif 'hosts' in token_dict:
            token_dict['hosts'] = nodeset(token_dict['hosts'])
            self._add_hosts(**token_dict)

        elif 'category' in token_dict:
            self._add_category(**token_dict)

        else:  # pragma: no cover - this should never happen
            raise InvalidQueryError(
                "No valid key found in token, one of bool|hosts|category expected: {token}".format(token=token_dict))

    def _get_resource_query(self, key, value=None, operator='='):  # pylint: disable=no-self-use
        """Build a resource query based on the parameters, resolving the special cases for ``%params`` and ``@field``.

        Arguments:
            key (str): the key of the resource.
            value (str, optional): the value to match, if not specified the key itself will be matched.
            operator (str, optional): the comparison operator to use, one of :py:const:`OPERATORS`.

        Returns:
            str: the resource query.

        Raises:
            cumin.backends.InvalidQueryError: on invalid combinations of parameters.

        """
        if all(char in key for char in ('%', '@')):
            raise InvalidQueryError(("Resource key cannot contain both '%' (query a resource's parameter) and '@' "
                                     "(query a  resource's field)"))

        if '%' in key:
            # Querying a specific parameter of the resource
            if operator == '~' and self.api_version == 3:
                raise InvalidQueryError('Regex operations are not supported in PuppetDB API v3 for resource parameters')
            key, param = key.split('%', 1)
            query_part = ', ["{op}", ["parameter", "{param}"], {value}]'.format(op=operator, param=param, value=value)

        elif '@' in key:
            # Querying a specific field of the resource
            key, field = key.split('@', 1)
            query_part = ', ["{op}", "{field}", {value}]'.format(op=operator, field=field, value=value)

        elif value is None:
            # Querying a specific resource type
            query_part = ''

        else:
            # Querying a specific resource title
            if key.lower() == 'class' and operator != '~':
                value = value.capwords('::')  # Auto ucfirst the class title
            query_part = ', ["{op}", "title", {value}]'.format(op=operator, value=value)
        query = '["and", ["=", "type", "{type}"]{query_part}]'.format(type=capwords(key, '::'), query_part=query_part)

        return query

    def _get_special_resource_query(self, category, key, value, operator):
        """Build a query for Roles and Profiles, resolving the special cases for ``%params`` and ``@field``.

        Arguments:
            category (str): the category of the token, one of :py:data:`category_prefixes` keys.
            key (str): the key of the resource to use as a suffix for the Class title matching.
            value (str, optional): the value to match in case ``%params`` or ``@field`` is specified.
            operator (str, optional): the comparison operator to use if there is a value, one of :py:const:`OPERATORS`.

        Returns:
            str: the resource query.

        Raises:
            cumin.backends.InvalidQueryError: on invalid combinations of parameters.

        """
        if all(char in key for char in ('%', '@')):
            raise InvalidQueryError(("Resource key cannot contain both '%' (query a resource's parameter) and '@' "
                                     "(query a  resource's field)"))
        if '%' in key:
            special = '%'
            key, param = key.split('%')
        elif '@' in key:
            special = '@'
            key, param = key.split('@')
        else:
            special = None
            if value is not None:
                raise InvalidQueryError(("Invalid query of the form '{category}:key = value'. The matching of a value "
                                         "is accepted only when using %param or @field.").format(category=category))

        if self.category_prefixes[category]:
            title = ParsedString('{prefix}::{key}'.format(prefix=self.category_prefixes[category], key=key), True)
        else:
            title = ParsedString(key, True)

        query = self._get_resource_query('Class', title, '=')

        if special is not None:
            param_query = self._get_resource_query(''.join(('Class', special, param)), value, operator)
            query = '["and", {query}, {param_query}]'.format(query=query, param_query=param_query)

        return query

    def _get_query_string(self, group):
        """Recursively build and return the PuppetDB query string.

        Arguments:
            group (dict): a dictionary with the grouped tokens.

        Returns:
            str: the query string for the PuppetDB API.

        """
        if group['bool']:
            query = '["{bool}", '.format(bool=group['bool'])
        else:
            query = ''

        last_index = len(group['tokens'])
        for i, token in enumerate(group['tokens']):
            if isinstance(token, dict):
                query += self._get_query_string(group=token)
            else:
                query += token

            if i < last_index - 1:
                query += ', '

        if group['bool']:
            query += ']'

        return query

    def _add_bool(self, bool_op):
        """Add a boolean AND or OR query block to the query and validate logic.

        Arguments:
            bool_op (str): the boolean operator to add to the query: ``and``, ``or``.

        Raises:
            cumin.backends.InvalidQueryError: if an invalid boolean operator was found.

        """
        if self.current_group['bool'] is None:
            self.current_group['bool'] = bool_op
        elif self.current_group['bool'] == bool_op:
            return
        else:
            raise InvalidQueryError("Got unexpected '{bool}' boolean operator, current operator was '{current}'".format(
                bool=bool_op, current=self.current_group['bool']))

    def _api_call(self, query):
        """Execute a query to PuppetDB API and return the parsed JSON.

        Arguments:
            query (str): the query parameter to send to the PuppetDB API.

        Raises:
            requests.HTTPError: if the PuppetDB API call fails.

        """
        if self.api_version == 3:
            resources = requests.get(self.url + self.endpoint, params={'query': query}, verify=True)
        else:
            resources = requests.post(self.url + self.endpoint, json={'query': query}, verify=True)

        resources.raise_for_status()
        return resources.json()


GRAMMAR_PREFIX = 'P'
""":py:class:`str`: the prefix associate to this grammar, to register this backend into the general grammar.
Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""

query_class = PuppetDBQuery  # pylint: disable=invalid-name
"""Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""



"""Direct backend."""
import pyparsing as pp

from cumin import nodeset_fromlist
from cumin.backends import BaseQueryAggregator, InvalidQueryError


def grammar():
    """Define the query grammar.

    Backus-Naur form (BNF) of the grammar::

        <grammar> ::= <item> | <item> <boolean> <grammar>
           <item> ::= <hosts> | "(" <grammar> ")"
        <boolean> ::= "and not" | "and" | "xor" | "or"

    Given that the pyparsing library defines the grammar in a BNF-like style, for the details of the tokens not
    specified above check directly the source code.

    Returns:
        pyparsing.ParserElement: the grammar parser.

    """
    # Boolean operators
    boolean = (pp.CaselessKeyword('and not').leaveWhitespace() | pp.CaselessKeyword('and')
               | pp.CaselessKeyword('xor') | pp.CaselessKeyword('or'))('bool')

    # Parentheses
    lpar = pp.Literal('(')('open_subgroup')
    rpar = pp.Literal(')')('close_subgroup')

    # Hosts selection: clustershell (,!&^[]) syntax is allowed: host10[10-42].domain
    hosts = (~(boolean) + pp.Word(pp.alphanums + '-_.,!&^[]'))('hosts')

    # Final grammar, see the docstring for its BNF based on the tokens defined above
    # Groups are used to split the parsed results for an easy access
    full_grammar = pp.Forward()
    item = hosts | lpar + full_grammar + rpar
    full_grammar << pp.Group(item) + pp.ZeroOrMore(pp.Group(boolean + item))  # pylint: disable=expression-not-assigned

    return full_grammar


class DirectQuery(BaseQueryAggregator):
    """DirectQuery query builder.

    The `direct` backend allow to use Cumin without any external dependency for the hosts selection.
    It allow to write arbitrarily complex queries with subgroups and boolean operators, but each item must be either the
    hostname itself, or the using host expansion using the powerful :py:class:`ClusterShell.NodeSet.NodeSet` syntax.

    The typical usage for the `direct` backend is as a reliable alternative in cases in which the primary host
    selection mechanism is not working and also for testing the transports without any external backend dependency.

    Some query examples:

    * Simple selection: ``host1.domain``
    * ClusterShell syntax for hosts expansion: ``host10[10-42].domain,host2010.other-domain``
    * A complex selection:
      ``host100[1-5].domain or (host10[30-40].domain and (host10[10-42].domain and not host33.domain))``
    """

    grammar = grammar()
    """:py:class:`pyparsing.ParserElement`: load the grammar parser only once in a singleton-like way."""

    def _parse_token(self, token):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQueryAggregator._parse_token`.
        """
        if not isinstance(token, pp.ParseResults):  # pragma: no cover - this should never happen
            raise InvalidQueryError('Expecting ParseResults object, got {type}: {token}'.format(
                type=type(token), token=token))

        token_dict = token.asDict()
        self.logger.trace('Token is: %s | %s', token_dict, token)

        if 'hosts' in token_dict:
            element = self._get_stack_element()
            element['hosts'] = nodeset_fromlist(token_dict['hosts'])
            if 'bool' in token_dict:
                element['bool'] = token_dict['bool']
            self.stack_pointer['children'].append(element)
        elif 'open_subgroup' in token_dict and 'close_subgroup' in token_dict:
            self._open_subgroup()
            if 'bool' in token_dict:
                self.stack_pointer['bool'] = token_dict['bool']
            for subtoken in token:
                if isinstance(subtoken, str):  # Grammar literals, boolean operators and parentheses
                    continue
                self._parse_token(subtoken)
            self._close_subgroup()
        else:  # pragma: no cover - this should never happen
            raise InvalidQueryError('Got unexpected token: {token}'.format(token=token))


GRAMMAR_PREFIX = 'D'
""":py:class:`str`: the prefix associate to this grammar, to register this backend into the general grammar.
Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""

query_class = DirectQuery  # pylint: disable=invalid-name
"""Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""



"""Cumin package tests."""
# pylint: disable=invalid-name
import importlib
import logging
import os

import pytest

from ClusterShell.NodeSet import NodeSet

import cumin

from cumin.tests import get_fixture_path


def test_config_class_valid():
    """Should return the config. Multiple Config with the same path should return the same object."""
    config_file = get_fixture_path(os.path.join('config', 'valid', 'config.yaml'))
    config1 = cumin.Config(config=config_file)
    assert 'log_file' in config1
    config2 = cumin.Config(config=config_file)
    assert config1 is config2


def test_config_class_empty():
    """An empty dictionary is returned if the configuration is empty."""
    config = cumin.Config(config=get_fixture_path(os.path.join('config', 'empty', 'config.yaml')))
    assert config == {}


def test_config_class_invalid():
    """A CuminError is raised if the configuration cannot be parsed."""
    with pytest.raises(cumin.CuminError, match='Unable to parse configuration file'):
        cumin.Config(config=get_fixture_path(os.path.join('config', 'invalid', 'config.yaml')))


def test_config_class_valid_with_aliases():
    """Should return the config including the aliases."""
    config = cumin.Config(config=get_fixture_path(os.path.join('config', 'valid_with_aliases', 'config.yaml')))
    assert 'log_file' in config
    assert 'aliases' in config
    assert 'role1' in config['aliases']
    assert config['aliases']['role1'] == 'P{R:Class = Role::Role1}'
    assert 'group1' in config['aliases']
    assert config['aliases']['group1'] == 'D{host10[10-22].example.org}'


def test_config_class_empty_aliases():
    """The configuration is loaded also if the aliases file is empty."""
    config = cumin.Config(config=get_fixture_path(os.path.join('config', 'valid_with_empty_aliases', 'config.yaml')))
    assert 'log_file' in config
    assert 'aliases' in config
    assert config['aliases'] == {}


def test_config_class_invalid_aliases():
    """A CuminError is raised if one of the backend aliases is invalid."""
    with pytest.raises(cumin.CuminError, match='Unable to parse configuration file'):
        cumin.Config(config=get_fixture_path(os.path.join('config', 'valid_with_invalid_aliases', 'config.yaml')))


def test_parse_config_ok():
    """The configuration file is properly parsed and accessible."""
    config = cumin.parse_config(get_fixture_path(os.path.join('config', 'valid', 'config.yaml')))
    assert 'log_file' in config


def test_parse_config_non_existent():
    """A CuminError is raised if the configuration file is not available."""
    with pytest.raises(cumin.CuminError, match='Unable to read configuration file'):
        cumin.parse_config('not_existent_config.yaml')


def test_parse_config_invalid():
    """A CuminError is raised if the configuration cannot be parsed."""
    with pytest.raises(cumin.CuminError, match='Unable to parse configuration file'):
        cumin.parse_config(get_fixture_path(os.path.join('config', 'invalid', 'config.yaml')))


def test_parse_config_empty():
    """An empty dictionary is returned if the configuration is empty."""
    config = cumin.parse_config(get_fixture_path(os.path.join('config', 'empty', 'config.yaml')))
    assert config == {}


def test_trace_logging_level_conflict():
    """If the logging level for trace is already registered, should raise CuminError."""
    importlib.reload(logging)  # Avoid conflict given the singleton nature of this module
    logging.addLevelName(cumin.LOGGING_TRACE_LEVEL_NUMBER, 'CONFLICT')
    match = 'Unable to set custom logging for trace'
    try:  # pytest.raises doesn't catch the reload exception
        importlib.reload(cumin)
    except cumin.CuminError as e:
        assert str(e).startswith(match)
    else:
        raise AssertionError("Failed: DID NOT RAISE {exc} matching '{match}'".format(
            exc=cumin.CuminError, match=match))


def test_trace_logging_level_existing_same():
    """If the custom logging level is registered on the same level, it should use it and add a trace method."""
    importlib.reload(logging)  # Avoid conflict given the singleton nature of this module
    logging.addLevelName(cumin.LOGGING_TRACE_LEVEL_NUMBER, cumin.LOGGING_TRACE_LEVEL_NAME)
    assert not hasattr(logging.Logger, 'trace')
    importlib.reload(cumin)
    assert logging.getLevelName(cumin.LOGGING_TRACE_LEVEL_NUMBER) == cumin.LOGGING_TRACE_LEVEL_NAME
    assert logging.getLevelName(cumin.LOGGING_TRACE_LEVEL_NAME) == cumin.LOGGING_TRACE_LEVEL_NUMBER
    assert hasattr(logging.Logger, 'trace')


def test_trace_logging_level_existing_different():
    """If the custom logging level is registered on a different level, it should use it and add a trace method."""
    importlib.reload(logging)  # Avoid conflict given the singleton nature of this module
    logging.addLevelName(cumin.LOGGING_TRACE_LEVEL_NUMBER - 1, cumin.LOGGING_TRACE_LEVEL_NAME)
    assert not hasattr(logging.Logger, 'trace')
    importlib.reload(cumin)
    assert logging.getLevelName(cumin.LOGGING_TRACE_LEVEL_NAME) == cumin.LOGGING_TRACE_LEVEL_NUMBER - 1
    assert logging.getLevelName(cumin.LOGGING_TRACE_LEVEL_NUMBER) != cumin.LOGGING_TRACE_LEVEL_NAME
    assert hasattr(logging.Logger, 'trace')


def test_trace_logging_method_existing():
    """If there is already a trace method registered, it should use it without problems adding the level."""
    importlib.reload(logging)  # Avoid conflict given the singleton nature of this module
    logging.Logger.trace = cumin.trace
    importlib.reload(cumin)
    assert logging.getLevelName(cumin.LOGGING_TRACE_LEVEL_NUMBER) == cumin.LOGGING_TRACE_LEVEL_NAME
    assert logging.getLevelName(cumin.LOGGING_TRACE_LEVEL_NAME) == cumin.LOGGING_TRACE_LEVEL_NUMBER
    assert hasattr(logging.Logger, 'trace')


def test_nodeset():
    """Calling nodeset() should return an instance of ClusterShell NodeSet with no resolver."""
    nodeset = cumin.nodeset('node[1-2]')
    assert isinstance(nodeset, NodeSet)
    assert nodeset == NodeSet('node[1-2]')
    assert nodeset._resolver is None  # pylint: disable=protected-access


def test_nodeset_empty():
    """Calling nodeset() without parameter should return an instance of ClusterShell NodeSet with no resolver."""
    nodeset = cumin.nodeset()
    assert isinstance(nodeset, NodeSet)
    assert nodeset == NodeSet()
    assert nodeset._resolver is None  # pylint: disable=protected-access


def test_nodeset_fromlist():
    """Calling nodeset_fromlist() should return an instance of ClusterShell NodeSet with no resolver."""
    nodeset = cumin.nodeset_fromlist(['node1', 'node2'])
    assert isinstance(nodeset, NodeSet)
    assert nodeset == NodeSet('node[1-2]')
    assert nodeset._resolver is None  # pylint: disable=protected-access


def test_nodeset_fromlist_empty():
    """Calling nodeset_fromlist() with empty list should return an instance of ClusterShell NodeSet with no resolver."""
    nodeset = cumin.nodeset_fromlist([])
    assert isinstance(nodeset, NodeSet)
    assert nodeset == NodeSet()
    assert nodeset._resolver is None  # pylint: disable=protected-access



"""External backends package for testing."""
import pyparsing as pp

from cumin import nodeset
from cumin.backends import BaseQuery


def grammar():
    """Define the query grammar for the external backend used for testing."""
    # Hosts selection: clustershell (,!&^[]) syntax is allowed: host10[10-42].domain
    hosts = pp.Word(pp.alphanums + '-_.,!&^[]')('hosts')

    # Final grammar, see the docstring for its BNF based on the tokens defined above
    # Groups are used to split the parsed results for an easy access
    full_grammar = pp.Forward()
    full_grammar << pp.Group(hosts) + pp.ZeroOrMore(pp.Group(hosts))  # pylint: disable=expression-not-assigned

    return full_grammar


class ExternalBackendQuery(BaseQuery):
    """External backend test query class."""

    grammar = grammar()
    """:py:class:`pyparsing.ParserElement`: load the grammar parser only once in a singleton-like way."""

    def __init__(self, config):
        """Query constructor for the test external backend.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery.__init__`.

        """
        super().__init__(config)
        self.hosts = nodeset()

    def _execute(self):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._execute`.

        Returns:
            ClusterShell.NodeSet.NodeSet: with the FQDNs of the matching hosts.

        """
        return self.hosts

    def _parse_token(self, token):
        """Concrete implementation of parent abstract method.

        :Parameters:
            according to parent :py:meth:`cumin.backends.BaseQuery._parse_token`.
        """
        if isinstance(token, str):
            return

        token_dict = token.asDict()
        self.hosts |= nodeset(token_dict['hosts'])



"""Query grammar definition."""
import importlib
import pkgutil

from collections import namedtuple

import pyparsing as pp

from cumin import backends, CuminError


INTERNAL_BACKEND_PREFIX = 'cumin.backends'
""":py:class:`str` with the prefix for built-in backends."""


Backend = namedtuple('Backend', ['keyword', 'name', 'cls'])
""":py:func:`collections.namedtuple` that define a Backend object.

Keyword Arguments:
    keyword (str): The backend keyword to be used in the grammar.
    name (str): The backend name.
    cls (BaseQuery): The backend class object.
"""


def get_registered_backends(external=()):
    """Get a mapping of all the registered backends with their keyword.

    Arguments:
        external (list, tuple, optional): external backend modules to register.

    Returns:
        dict: A dictionary with a ``{keyword: Backend object}`` mapping for each available backend.

    Raises:
        cumin.CuminError: If unable to register a backend.

    """
    available_backends = {}
    backend_names = ['{prefix}.{backend}'.format(prefix=INTERNAL_BACKEND_PREFIX, backend=name)
                     for _, name, ispkg in pkgutil.iter_modules(backends.__path__) if not ispkg]

    for name in backend_names + list(external):
        keyword, backend = _import_backend(name, available_backends)
        if keyword is not None and backend is not None:
            available_backends[keyword] = backend

    return available_backends


def grammar(backend_keys):
    """Define the main multi-query grammar.

    Cumin provides a user-friendly generic query language that allows to combine the results of subqueries for multiple
    backends:

    * Each query part can be composed with the others using boolean operators ``and``, ``or``, ``and not``, ``xor``.
    * Multiple query parts can be grouped together with parentheses ``(``, ``)``.
    * Specific backend query ``I{backend-specific query syntax}``, where ``I`` is an identifier for the specific
      backend.
    * Alias replacement, according to aliases defined in the configuration file ``A:group1``.
    * The identifier ``A`` is reserved for the aliases replacement and cannot be used to identify a backend.
    * A complex query example: ``(D{host1 or host2} and (P{R:Class = Role::MyClass} and not A:group1)) or D{host3}``

    Backus-Naur form (BNF) of the grammar::

              <grammar> ::= <item> | <item> <boolean> <grammar>
                 <item> ::= <backend_query> | <alias> | "(" <grammar> ")"
        <backend_query> ::= <backend> "{" <query> "}"
                <alias> ::= A:<alias_name>
              <boolean> ::= "and not" | "and" | "xor" | "or"

    Given that the pyparsing library defines the grammar in a BNF-like style, for the details of the tokens not
    specified above check directly the source code.

    Arguments:
        backend_keys (list): list of the GRAMMAR_PREFIX for each registered backend.

    Returns:
        pyparsing.ParserElement: the grammar parser.

    """
    # Boolean operators
    boolean = (pp.CaselessKeyword('and not').leaveWhitespace() | pp.CaselessKeyword('and')
               | pp.CaselessKeyword('xor') | pp.CaselessKeyword('or'))('bool')

    # Parentheses
    lpar = pp.Literal('(')('open_subgroup')
    rpar = pp.Literal(')')('close_subgroup')

    # Backend query: P{PuppetDB specific query}
    query_start = pp.Combine(pp.oneOf(backend_keys, caseless=True)('backend') + pp.Literal('{'))
    query_end = pp.Literal('}')
    # Allow the backend specific query to use the end_query token as well, as long as it's in a quoted string
    # and fail if there is a query_start token before the first query_end is reached
    query = pp.SkipTo(query_end, ignore=pp.quotedString, failOn=query_start)('query')
    backend_query = pp.Combine(query_start + query + query_end)

    # Alias
    alias = pp.Combine(pp.CaselessKeyword('A') + ':' + pp.Word(pp.alphanums + '-_.+')('alias'))

    # Final grammar, see the docstring for its BNF based on the tokens defined above
    # Group are used to have an easy dictionary access to the parsed results
    full_grammar = pp.Forward()
    item = backend_query | alias | lpar + full_grammar + rpar
    full_grammar << pp.Group(item) + pp.ZeroOrMore(pp.Group(boolean + item))  # pylint: disable=expression-not-assigned

    return full_grammar


def _import_backend(module, available_backends):
    """Dynamically import a backend for Cumin and validate it.

    Arguments:
        module (str): the full module name of the backend to register. Must be importable from Python ``PATH``.
        available_backends (dict): dictionary with a ``{keyword: Backend object}`` mapping for all registered backends.

    Returns:
        tuple: with two elements: ``(keyword, Backend object)`` of the imported backend.

    """
    try:
        backend = importlib.import_module(module)
    except ImportError as e:
        if not module.startswith(INTERNAL_BACKEND_PREFIX):
            raise CuminError("Unable to import backend '{module}': {e}".format(module=module, e=e))

        return (None, None)  # Internal backend not available, are all the dependencies installed?

    name = module.split('.')[-1]
    message = "Unable to register backend '{name}' in module '{module}'".format(name=name, module=module)
    try:
        keyword = backend.GRAMMAR_PREFIX
    except AttributeError:
        raise CuminError('{message}: GRAMMAR_PREFIX module attribute not found'.format(message=message))

    if keyword in available_backends:
        raise CuminError(("{message}: keyword '{key}' already registered: {backends}").format(
            message=message, key=keyword, backends=available_backends))

    try:
        class_obj = backend.query_class
    except AttributeError:
        raise CuminError('{message}: query_class module attribute not found'.format(message=message))

    if not issubclass(class_obj, backends.BaseQuery):
        raise CuminError('{message}: query_class module attribute is not a subclass of cumin.backends.BaseQuery')

    return (keyword, Backend(name=name, keyword=keyword, cls=class_obj))
